from enum import Enum
from re import compile as regex
from typing import Any, Dict, List, Optional, Tuple, Type, Union

from fideslang.validation import FidesKey
from pydantic import BaseModel, Extra, root_validator

from fides.api.ops.models.privacy_request import CheckpointActionRequired
from fides.api.ops.schemas import Msg


class MessagingMethod(Enum):
    """Enum for messaging method"""

    EMAIL = "email"
    SMS = "sms"


class MessagingServiceType(Enum):
    """Enum for messaging service type. Upper-cased in the database"""

    MAILGUN = "MAILGUN"

    TWILIO_TEXT = "TWILIO_TEXT"
    TWILIO_EMAIL = "TWILIO_EMAIL"


EMAIL_MESSAGING_SERVICES: Tuple[str, ...] = (
    MessagingServiceType.MAILGUN.value,
    MessagingServiceType.TWILIO_EMAIL.value,
)
SMS_MESSAGING_SERVICES: Tuple[str, ...] = (MessagingServiceType.TWILIO_TEXT.value,)


class MessagingActionType(str, Enum):
    """Enum for messaging action type"""

    # verify email upon acct creation
    CONSENT_REQUEST = "consent_request"
    SUBJECT_IDENTITY_VERIFICATION = "subject_identity_verification"
    MESSAGE_ERASURE_REQUEST_FULFILLMENT = "message_erasure_fulfillment"
    PRIVACY_REQUEST_ERROR_NOTIFICATION = "privacy_request_error_notification"
    PRIVACY_REQUEST_RECEIPT = "privacy_request_receipt"
    PRIVACY_REQUEST_COMPLETE_ACCESS = "privacy_request_complete_access"
    PRIVACY_REQUEST_COMPLETE_DELETION = "privacy_request_complete_deletion"
    PRIVACY_REQUEST_REVIEW_DENY = "privacy_request_review_deny"
    PRIVACY_REQUEST_REVIEW_APPROVE = "privacy_request_review_approve"


class ErrorNotificaitonBodyParams(BaseModel):
    """Body params required for privacy request error notifications."""

    unsent_errors: int


class SubjectIdentityVerificationBodyParams(BaseModel):
    """Body params required for subject identity verification email/sms template"""

    verification_code: str
    verification_code_ttl_seconds: int

    def get_verification_code_ttl_minutes(self) -> int:
        """returns verification_code_ttl_seconds in minutes"""
        if self.verification_code_ttl_seconds < 60:
            return 0
        return self.verification_code_ttl_seconds // 60


class RequestReceiptBodyParams(BaseModel):
    """Body params required for privacy request receipt template"""

    request_types: List[str]


class AccessRequestCompleteBodyParams(BaseModel):
    """Body params required for privacy request completion access template"""

    download_links: List[str]


class RequestReviewDenyBodyParams(BaseModel):
    """Body params required for privacy request review deny template"""

    rejection_reason: Optional[str]


class FidesopsMessage(
    BaseModel,
    smart_union=True,
    arbitrary_types_allowed=True,
):
    """A mapping of action_type to body_params"""

    action_type: MessagingActionType
    body_params: Optional[
        Union[
            SubjectIdentityVerificationBodyParams,
            RequestReceiptBodyParams,
            RequestReviewDenyBodyParams,
            AccessRequestCompleteBodyParams,
            List[CheckpointActionRequired],
        ]
    ]


class EmailForActionType(BaseModel):
    """Email details that depend on action type"""

    subject: str
    body: str


class MessagingServiceDetails(Enum):
    """Enum for messaging service details"""

    # Mailgun
    IS_EU_DOMAIN = "is_eu_domain"
    API_VERSION = "api_version"
    DOMAIN = "domain"

    # Twilio Email
    TWILIO_EMAIL_FROM = "twilio_email_from"


class MessagingServiceDetailsMailgun(BaseModel):
    """The details required to represent a Mailgun email configuration."""

    is_eu_domain: Optional[bool] = False
    api_version: Optional[str] = "v3"
    domain: str

    class Config:
        """Restrict adding other fields through this schema."""

        extra = Extra.forbid


class MessagingServiceDetailsTwilioEmail(BaseModel):
    """The details required to represent a Twilio email configuration."""

    twilio_email_from: str

    class Config:
        """Restrict adding other fields through this schema."""

        extra = Extra.forbid


class MessagingServiceSecrets(Enum):
    """Enum for message service secrets"""

    # Mailgun
    MAILGUN_API_KEY = "mailgun_api_key"

    # Twilio SMS
    TWILIO_ACCOUNT_SID = "twilio_account_sid"
    TWILIO_AUTH_TOKEN = "twilio_auth_token"
    TWILIO_MESSAGING_SERVICE_SID = "twilio_messaging_service_sid"
    TWILIO_SENDER_PHONE_NUMBER = "twilio_sender_phone_number"

    # Twilio Sendgrid/Email
    TWILIO_API_KEY = "twilio_api_key"


class MessagingServiceSecretsMailgun(BaseModel):
    """The secrets required to connect to mailgun."""

    mailgun_api_key: str

    class Config:
        """Restrict adding other fields through this schema."""

        extra = Extra.forbid


class MessagingServiceSecretsTwilioSMS(BaseModel):
    """The secrets required to connect to twilio SMS."""

    twilio_account_sid: str
    twilio_auth_token: str
    twilio_messaging_service_sid: Optional[str]
    twilio_sender_phone_number: Optional[str]

    class Config:
        """Restrict adding other fields through this schema."""

        extra = Extra.forbid

    @root_validator
    def validate_fields(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        sender_phone = values.get("twilio_sender_phone_number")
        if not values.get("twilio_messaging_service_sid") and not sender_phone:
            raise ValueError(
                "Either the twilio_messaging_service_sid or the twilio_sender_phone_number should be supplied."
            )
        if sender_phone:
            pattern = regex(r"^\+\d+$")
            if not pattern.search(sender_phone):
                raise ValueError(
                    "Sender phone number must include country code, formatted like +15558675309"
                )
        return values


class MessagingServiceSecretsTwilioEmail(BaseModel):
    """The secrets required to connect to twilio email."""

    twilio_api_key: str

    class Config:
        """Restrict adding other fields through this schema."""

        extra = Extra.forbid


class MessagingConfigRequest(BaseModel):
    """Messaging Config Request Schema"""

    name: str
    key: Optional[FidesKey]
    service_type: MessagingServiceType
    details: Optional[
        Union[MessagingServiceDetailsMailgun, MessagingServiceDetailsTwilioEmail]
    ]

    class Config:
        use_enum_values = False
        orm_mode = True

    @root_validator(pre=True)
    def validate_fields(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        service_type_pre = values.get("service_type")
        if service_type_pre:
            # uppercase to match enums in database
            values["service_type"] = service_type_pre.upper()
            service_type: MessagingServiceType = values["service_type"]
            if service_type == MessagingServiceType.MAILGUN.value:
                cls._validate_details_schema(
                    values=values, schema=MessagingServiceDetailsMailgun
                )
            if service_type == MessagingServiceType.TWILIO_EMAIL.value:
                cls._validate_details_schema(
                    values=values, schema=MessagingServiceDetailsTwilioEmail
                )
        return values

    @staticmethod
    def _validate_details_schema(
        values: Dict[str, Any],
        schema: Union[
            Type[MessagingServiceDetailsMailgun],
            Type[MessagingServiceDetailsTwilioEmail],
        ],
    ) -> None:
        if not values.get("details"):
            raise ValueError("Messaging config must include details")
        schema.validate(values.get("details"))


class MessagingConfigResponse(BaseModel):
    """Messaging Config Response Schema"""

    name: str
    key: FidesKey
    service_type: MessagingServiceType
    details: Optional[Dict[MessagingServiceDetails, Any]]

    class Config:
        orm_mode = True
        use_enum_values = True


SUPPORTED_MESSAGING_SERVICE_SECRETS = Union[
    MessagingServiceSecretsMailgun,
    MessagingServiceSecretsTwilioSMS,
    MessagingServiceSecretsTwilioEmail,
]


class MessagingConnectionTestStatus(Enum):
    """Enum for supplying statuses of validating credentials for a messaging Config"""

    succeeded = "succeeded"
    failed = "failed"
    skipped = "skipped"


class TestMessagingStatusMessage(Msg):
    """A schema for checking status of message config."""

    test_status: Optional[MessagingConnectionTestStatus] = None
    failure_reason: Optional[str] = None



"""Add fidesctl_meta field to dataset

Revision ID: be432bd23596
Revises: adad3be6af08
Create Date: 2022-04-01 00:15:07.569104

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "be432bd23596"
down_revision = "a2b9b156ebaf"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("datasets", sa.Column("fidesctl_meta", sa.JSON(), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("datasets", "fidesctl_meta")
    # ### end Alembic commands ###



"""Merge failed dsr and twilio

Revision ID: 58933b5cc6e8
Revises: 179f2bb623ae, 28108b17a99c
Create Date: 2022-11-14 21:26:49.027809

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "58933b5cc6e8"
down_revision = ("179f2bb623ae", "28108b17a99c")
branch_labels = None
depends_on = None


def upgrade():
    pass


def downgrade():
    pass



"""
Reusable utilities meant to make repetitive api-related tasks easier.
"""

from typing import Dict, List, Optional, Union

from fideslang import FidesModel
from fideslang.parse import parse_dict
from fideslang.validation import FidesKey
from requests import Response

from fides.core import api
from fides.core.utils import check_response_auth


def get_server_resources(
    url: str,
    resource_type: str,
    existing_keys: List[FidesKey],
    headers: Dict[str, str],
) -> List[FidesModel]:
    """
    Get a list of resources from the server that match the provided keys.

    If the resource does not exist on the server, an error will _not_ be thrown.
    Instead, an empty object will be stored and then filtered out.
    """
    server_resources: List[FidesModel] = list(
        filter(
            None,
            [
                get_server_resource(
                    url=url,
                    resource_type=resource_type,
                    resource_key=key,
                    headers=headers,
                )
                for key in existing_keys
            ],
        )
    )
    return server_resources


def get_server_resource(
    url: str,
    resource_type: str,
    resource_key: str,
    headers: Dict[str, str],
    raw: bool = False,
) -> Union[FidesModel, Dict]:
    """
    Attempt to get a given resource from the server.

    Returns {} if the object does not exist on the server.

    As we don't always have a way to attribute fides_keys to the
    right resource, this function helps check what resource
    a fides_key belongs to by allowing us to check without errors.
    """
    raw_server_response: Response = check_response_auth(
        api.get(
            url=url,
            resource_type=resource_type,
            resource_id=resource_key,
            headers=headers,
        )
    )

    server_resource: Optional[FidesModel] = (
        raw_server_response.json()
        if raw_server_response.status_code >= 200
        and raw_server_response.status_code <= 299
        else None
    )
    if not raw and server_resource:
        server_resource = parse_dict(
            resource_type=resource_type,
            resource=raw_server_response.json(),
            from_server=True,
        )

    return server_resource or {}


def list_server_resources(
    url: str,
    headers: Dict[str, str],
    resource_type: str,
    exclude_keys: List[str],
    raw: bool = False,
) -> Optional[Union[List[FidesModel], List[Dict]]]:
    """
    Get a list of resources from the server and return them as parsed objects.

    Returns an empty list if no resources are found or if the API returns an error.
    """
    response: Response = check_response_auth(
        api.ls(url=url, resource_type=resource_type, headers=headers)
    )
    server_resources = (
        [
            resource
            for resource in response.json()
            if isinstance(resource, dict) and resource["fides_key"] not in exclude_keys
        ]
        if response.status_code >= 200 and response.status_code <= 299
        else []
    )

    if not raw and server_resources:
        server_resources = [
            parse_dict(
                resource_type=resource_type,
                resource=resource_dict,
                from_server=True,
            )
            for resource_dict in server_resources
        ]

    return server_resources



"""another fidesops remote merge

Revision ID: 25adddf820a3
Revises: 7b81d34352e8, 97801300fedd
Create Date: 2022-08-24 10:40:25.875324

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "25adddf820a3"
down_revision = ("7b81d34352e8", "97801300fedd")
branch_labels = None
depends_on = None


def upgrade():
    pass


def downgrade():
    pass



import re
from datetime import datetime
from typing import Optional

from pydantic import validator

from fides.lib.cryptography.cryptographic_util import decode_password
from fides.lib.oauth.schemas.oauth import AccessToken
from fides.lib.schemas.base_class import BaseSchema


class PrivacyRequestReviewer(BaseSchema):
    """Data we can expose via the PrivacyRequest.reviewer relation"""

    id: str
    username: str


class UserCreate(BaseSchema):
    """Data required to create a FidesUser."""

    username: str
    password: str
    first_name: Optional[str]
    last_name: Optional[str]

    @validator("username")
    @classmethod
    def validate_username(cls, username: str) -> str:
        """Ensure password does not have spaces."""
        if " " in username:
            raise ValueError("Usernames cannot have spaces.")
        return username

    @validator("password")
    @classmethod
    def validate_password(cls, password: str) -> str:
        """Add some password requirements"""
        decoded_password = decode_password(password)

        if len(decoded_password) < 8:
            raise ValueError("Password must have at least eight characters.")
        if re.search("[0-9]", decoded_password) is None:
            raise ValueError("Password must have at least one number.")
        if re.search("[A-Z]", decoded_password) is None:
            raise ValueError("Password must have at least one capital letter.")
        if re.search("[a-z]", decoded_password) is None:
            raise ValueError("Password must have at least one lowercase letter.")
        if re.search(r"[\W]", decoded_password) is None:
            raise ValueError("Password must have at least one symbol.")

        return decoded_password


class UserCreateResponse(BaseSchema):
    """Response after creating a FidesUser"""

    id: str


class UserLogin(BaseSchema):
    """Similar to UserCreate except we do not need the extra validation on
    username and password.
    """

    username: str
    password: str

    @validator("password")
    @classmethod
    def validate_password(cls, password: str) -> str:
        """Convert b64 encoded password to normal string"""
        return decode_password(password)


class UserResponse(BaseSchema):
    """Response after requesting a User"""

    id: str
    username: str
    created_at: datetime
    first_name: Optional[str]
    last_name: Optional[str]


class UserLoginResponse(BaseSchema):
    """Similar to UserResponse except with an access token"""

    user_data: UserResponse
    token_data: AccessToken


class UserPasswordReset(BaseSchema):
    """Contains both old and new passwords when resetting a password"""

    old_password: str
    new_password: str


class UserForcePasswordReset(BaseSchema):
    """Only a new password, for the case where the user does not remember their password"""

    new_password: str


class UserUpdate(BaseSchema):
    """Data required to update a FidesopsUser"""

    first_name: Optional[str]
    last_name: Optional[str]



"""Contains the groups and setup for the CLI."""
from importlib.metadata import version
from platform import system

import click
from fideslog.sdk.python.client import AnalyticsClient

import fides
from fides.cli.utils import check_server
from fides.core.config import get_config
from fides.core.user import get_auth_header

from .commands.annotate import annotate
from .commands.core import evaluate, parse, pull, push
from .commands.crud import delete, get_resource, list_resources
from .commands.db import database
from .commands.export import export
from .commands.generate import generate
from .commands.scan import scan
from .commands.user import user
from .commands.util import deploy, init, status, webserver, worker
from .commands.view import view

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])
LOCAL_COMMANDS = [deploy, evaluate, generate, init, scan, parse, view, webserver]
LOCAL_COMMAND_DICT = {
    command.name or str(command): command for command in LOCAL_COMMANDS
}
API_COMMANDS = [
    annotate,
    database,
    delete,
    export,
    get_resource,
    list_resources,
    status,
    pull,
    push,
    worker,
    user,
]
API_COMMAND_DICT = {command.name or str(command): command for command in API_COMMANDS}
ALL_COMMANDS = API_COMMANDS + LOCAL_COMMANDS
SERVER_CHECK_COMMAND_NAMES = [
    command.name for command in API_COMMANDS if command.name not in ["status", "worker"]
]
VERSION = fides.__version__
APP = fides.__name__
PACKAGE = "ethyca-fides"


@click.group(
    context_settings=CONTEXT_SETTINGS,
    invoke_without_command=True,
    name="fides",
)
@click.version_option(version=VERSION)
@click.option(
    "--config-path",
    "-f",
    "config_path",
    default="",
    help="Path to a configuration file. Use 'fides view-config' to print the config. Not compatible with the 'fides webserver' subcommand.",
)
@click.option(
    "--local",
    is_flag=True,
    help="Run in 'local_mode'. This mode doesn't make API calls and can be used without the API server/database.",
)
@click.pass_context
def cli(ctx: click.Context, config_path: str, local: bool) -> None:
    """
    The parent group for the Fides CLI.
    """

    ctx.ensure_object(dict)
    config = get_config(config_path, verbose=True)
    try:
        config.user.auth_header = get_auth_header(verbose=False)
    except SystemExit:
        pass

    # Dyanmically add commands to the CLI
    cli.commands = LOCAL_COMMAND_DICT

    if not (local or config.cli.local_mode):
        config.cli.local_mode = False
        cli.commands = {**cli.commands, **API_COMMAND_DICT}
    else:
        config.cli.local_mode = True

    # Run the help command if no subcommand is passed
    if not ctx.invoked_subcommand:
        click.echo(cli.get_help(ctx))

    # Check the server health and version if an API command is invoked
    if ctx.invoked_subcommand in SERVER_CHECK_COMMAND_NAMES:
        check_server(VERSION, str(config.cli.server_url), quiet=True)

    # Analytics requires explicit opt-in
    no_analytics = config.user.analytics_opt_out
    if not no_analytics:
        ctx.meta["ANALYTICS_CLIENT"] = AnalyticsClient(
            client_id=config.cli.analytics_id,
            developer_mode=config.test_mode,
            os=system(),
            product_name=APP + "-cli",
            production_version=version(PACKAGE),
        )

    # Setting the config context after all mutations
    ctx.obj["CONFIG"] = config


# Add all commands here before dynamically checking them in the CLI
for cli_command in ALL_COMMANDS:
    cli.add_command(cli_command)



from os import getenv

DEFAULT_CONFIG_PATH = ".fides/fides.toml"
DEFAULT_CONFIG_PATH_ENV_VAR = "FIDES__CONFIG_PATH"


def get_test_mode() -> bool:
    test_mode = getenv("FIDES__TEST_MODE", "").lower() == "true"
    return test_mode


def get_dev_mode() -> bool:
    dev_mode = getenv("FIDES__DEV_MODE", "").lower() == "true"
    return dev_mode


CONFIG_KEY_ALLOWLIST = {
    "cli": ["server_host", "server_port"],
    "user": ["analytics_opt_out"],
    "logging": ["level"],
    "database": [
        "server",
        "user",
        "port",
        "db",
        "test_db",
    ],
    "notifications": [
        "send_request_completion_notification",
        "send_request_receipt_notification",
        "send_request_review_notification",
        "notification_service_type",
    ],
    "redis": [
        "host",
        "port",
        "charset",
        "decode_responses",
        "default_ttl_seconds",
        "db_index",
    ],
    "security": [
        "cors_origins",
        "encoding",
        "oauth_access_token_expire_minutes",
    ],
    "execution": [
        "task_retry_count",
        "task_retry_delay",
        "task_retry_backoff",
        "require_manual_request_approval",
    ],
}



from typing import Any, Dict, List, Optional

from fideslang.validation import FidesKey

from fides.api.ops.models.policy import ActionType, DrpAction
from fides.api.ops.schemas.api import BulkResponse, BulkUpdateFailed
from fides.api.ops.schemas.base_class import BaseSchema
from fides.api.ops.schemas.storage.storage import StorageDestinationResponse
from fides.api.ops.util.data_category import DataCategory


class PolicyMaskingSpec(BaseSchema):
    """Models the masking strategy definition"""

    strategy: str
    configuration: Dict[str, Any]


class PolicyMaskingSpecResponse(BaseSchema):
    """
    The schema to use when returning a masking strategy via the API. This schema omits other
    potentially sensitive fields in the masking configuration, for example the encryption
    algorithm.
    """

    strategy: str


class RuleTarget(BaseSchema):
    """An external representation of a Rule's target DataCategory within a Fidesops Policy"""

    name: Optional[str]
    key: Optional[FidesKey]
    data_category: DataCategory  # type: ignore

    class Config:
        """Populate models with the raw value of enum fields, rather than the enum itself"""

        use_enum_values = True


class RuleBase(BaseSchema):
    """An external representation of a Rule within a Fidesops Policy"""

    name: str
    key: Optional[FidesKey]
    action_type: ActionType

    class Config:
        """Populate models with the raw value of enum fields, rather than the enum itself"""

        use_enum_values = True


class RuleCreate(RuleBase):
    """
    The schema to use when creating a Rule. This schema accepts a storage_destination_key
    over a composite object.
    """

    storage_destination_key: Optional[FidesKey]
    masking_strategy: Optional[PolicyMaskingSpec]


class RuleResponse(RuleBase):
    """
    The schema to use when returning a Rule via the API. This schema uses a censored version
    of the `PolicyMaskingSpec` that omits the configuration to avoid exposing secrets.
    """

    storage_destination: Optional[StorageDestinationResponse]
    masking_strategy: Optional[PolicyMaskingSpecResponse]


class RuleResponseWithTargets(RuleBase):
    """
    The schema to use when returning a Rule via the API and when including the Rule's targets
    is desired. This schema uses a censored version of the `PolicyMaskingSpec` that omits the
    configuration to avoid exposing secrets.
    """

    targets: Optional[List[RuleTarget]]
    storage_destination: Optional[StorageDestinationResponse]
    masking_strategy: Optional[PolicyMaskingSpecResponse]


class Rule(RuleBase):
    """A representation of a Rule that features all storage destination data."""

    storage_destination: Optional[StorageDestinationResponse]
    masking_strategy: Optional[PolicyMaskingSpec]


class Policy(BaseSchema):
    """An external representation of a Fidesops Policy"""

    name: str
    key: Optional[FidesKey]
    drp_action: Optional[DrpAction]
    execution_timeframe: Optional[int]

    class Config:
        """Populate models with the raw value of enum fields, rather than the enum itself"""

        use_enum_values = True
        orm_mode = True


class PolicyResponse(Policy):
    """A holistic view of a Policy record, including all foreign keys by default."""

    rules: Optional[List[RuleResponse]]
    drp_action: Optional[DrpAction]


class BulkPutRuleTargetResponse(BulkResponse):
    """Schema with mixed success/failure responses for Bulk Create/Update of RuleTarget responses."""

    succeeded: List[RuleTarget]
    failed: List[BulkUpdateFailed]


class BulkPutRuleResponse(BulkResponse):
    """Schema with mixed success/failure responses for Bulk Create/Update of Rule responses."""

    succeeded: List[RuleResponse]
    failed: List[BulkUpdateFailed]


class BulkPutPolicyResponse(BulkResponse):
    """Schema with mixed success/failure responses for Bulk Create/Update of Policy responses."""

    succeeded: List[PolicyResponse]
    failed: List[BulkUpdateFailed]



from typing import List, Optional, Type

from fides.api.ops.schemas.masking.masking_configuration import (
    StringRewriteMaskingConfiguration,
)
from fides.api.ops.schemas.masking.masking_strategy_description import (
    MaskingStrategyConfigurationDescription,
    MaskingStrategyDescription,
)
from fides.api.ops.service.masking.strategy.format_preservation import (
    FormatPreservation,
)
from fides.api.ops.service.masking.strategy.masking_strategy import MaskingStrategy


class StringRewriteMaskingStrategy(MaskingStrategy):
    """Masks the values with a pre-determined value"""

    name = "string_rewrite"
    configuration_model = StringRewriteMaskingConfiguration

    def __init__(
        self,
        configuration: StringRewriteMaskingConfiguration,
    ):
        self.rewrite_value = configuration.rewrite_value
        self.format_preservation = configuration.format_preservation

    def mask(
        self, values: Optional[List[str]], request_id: Optional[str]
    ) -> Optional[List[str]]:
        """Replaces the value with the value specified in strategy spec. Returns None if input is
        None"""
        if values is None:
            return None
        masked_values: List[str] = []
        for _ in range(len(values)):
            if self.format_preservation is not None:
                formatter = FormatPreservation(self.format_preservation)
                masked_values.append(formatter.format(self.rewrite_value))
            else:
                masked_values.append(self.rewrite_value)
        return masked_values

    def secrets_required(self) -> bool:
        return False

    # MR Note - We will need a way to ensure that this does not fall out of date. Given that it
    # includes subjective instructions, this is not straightforward to automate
    @classmethod
    def get_description(cls: Type[MaskingStrategy]) -> MaskingStrategyDescription:
        return MaskingStrategyDescription(
            name=cls.name,
            description="Masks the input value with a default string value",
            configurations=[
                MaskingStrategyConfigurationDescription(
                    key="rewrite_value",
                    description="The string that will replace existing values",
                )
            ],
        )

    @staticmethod
    def data_type_supported(data_type: Optional[str]) -> bool:
        """Determines whether or not the given data type is supported by this masking strategy"""
        supported_data_types = {"string"}
        return data_type in supported_data_types



from typing import List, Optional

from fides.api.ops.schemas.base_class import NoValidationSchema
from fides.api.ops.schemas.connection_configuration.connection_secrets import (
    ConnectionConfigSecretsSchema,
)


class EmailSchema(ConnectionConfigSecretsSchema):
    """Schema to validate the secrets needed for the EmailConnector"""

    to_email: str
    test_email: Optional[str]  # Email to send a connection test email

    _required_components: List[str] = ["to_email"]


class EmailDocsSchema(EmailSchema, NoValidationSchema):
    """EmailDocsSchema Secrets Schema for API Docs"""



from typing import Any, Dict, List

from fides.api.ops.graph.traversal import TraversalNode
from fides.api.ops.models.connectionconfig import ConnectionConfig, ConnectionTestStatus
from fides.api.ops.models.policy import Policy
from fides.api.ops.models.privacy_request import PrivacyRequest
from fides.api.ops.service.connectors.base_connector import BaseConnector
from fides.api.ops.util.collection_util import Row


class ManualWebhookConnector(BaseConnector[None]):
    def query_config(self, node: TraversalNode) -> None:  # type: ignore
        """
        Not applicable for this connector type. Manual Webhooks are not run as part of the traversal.
        There will not be a node associated with the ManualWebhook.
        """
        return None

    def create_client(self) -> None:
        """Not needed because this connector involves a human performing some lookup step"""
        return None

    def close(self) -> None:
        """Not applicable for this connector type."""
        return None

    def test_connection(self) -> ConnectionTestStatus:
        """Very simple checks to verify that a ManualWebhook configuration exists"""
        connection_config: ConnectionConfig = self.configuration
        if not connection_config.access_manual_webhook:
            return ConnectionTestStatus.failed
        if not connection_config.access_manual_webhook.fields:
            return ConnectionTestStatus.failed
        return ConnectionTestStatus.succeeded

    def retrieve_data(  # type: ignore
        self,
        node: TraversalNode,
        policy: Policy,
        privacy_request: PrivacyRequest,
        input_data: Dict[str, List[Any]],
    ) -> None:
        """
        Not applicable for a manual webhook.  Manual webhooks are not called as part of the traversal.
        """
        return None

    def mask_data(  # type: ignore
        self,
        node: TraversalNode,
        policy: Policy,
        privacy_request: PrivacyRequest,
        rows: List[Row],
        input_data: Dict[str, List[Any]],
    ) -> None:
        """
        Not applicable for a manual webhook.  Manual webhooks are not called as part of the traversal.
        """
        return None



"""privacy request stopped

Revision ID: 27fe9da9d0f9
Revises: 1ff88b7bd579
Create Date: 2022-06-06 20:28:08.893135

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "27fe9da9d0f9"
down_revision = "1ff88b7bd579"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "privacyrequest",
        sa.Column("paused_at", sa.DateTime(timezone=True), nullable=True),
    )


def downgrade():
    op.drop_column("privacyrequest", "paused_at")



from enum import Enum
from typing import Any, Dict, List

from pydantic import ValidationError

from fides.api.ops.common_exceptions import NoSuchStrategyException
from fides.api.ops.common_exceptions import ValidationError as FidesopsValidationError
from fides.api.ops.schemas.saas.strategy_configuration import StrategyConfiguration
from fides.api.ops.service.authentication.authentication_strategy import (
    AuthenticationStrategy,
)
from fides.api.ops.service.authentication.authentication_strategy_basic import (
    BasicAuthenticationStrategy,
)
from fides.api.ops.service.authentication.authentication_strategy_bearer import (
    BearerAuthenticationStrategy,
)
from fides.api.ops.service.authentication.authentication_strategy_oauth2_authorization_code import (
    OAuth2AuthorizationCodeAuthenticationStrategy,
)
from fides.api.ops.service.authentication.authentication_strategy_oauth2_client_credentials import (
    OAuth2ClientCredentialsAuthenticationStrategy,
)
from fides.api.ops.service.authentication.authentication_strategy_query_param import (
    QueryParamAuthenticationStrategy,
)


class SupportedAuthenticationStrategies(Enum):
    """
    The supported strategies for authenticating against SaaS APIs.
    """

    basic = BasicAuthenticationStrategy
    bearer = BearerAuthenticationStrategy
    query_param = QueryParamAuthenticationStrategy
    oauth2_authorization_code = OAuth2AuthorizationCodeAuthenticationStrategy
    oauth2_client_credentials = OAuth2ClientCredentialsAuthenticationStrategy

    @classmethod
    def __contains__(cls, item: str) -> bool:
        try:
            cls[item]
        except KeyError:
            return False

        return True


def get_strategy(
    strategy_name: str,
    configuration: Dict[str, Any],
) -> AuthenticationStrategy:
    """
    Returns the strategy given the name and configuration.
    Raises NoSuchStrategyException if the strategy does not exist
    """
    if not SupportedAuthenticationStrategies.__contains__(strategy_name):
        valid_strategies = ", ".join(get_strategy_names())
        raise NoSuchStrategyException(
            f"Strategy '{strategy_name}' does not exist. Valid strategies are [{valid_strategies}]"
        )
    strategy = SupportedAuthenticationStrategies[strategy_name].value
    try:
        strategy_config: StrategyConfiguration = strategy.get_configuration_model()(
            **configuration
        )
        return strategy(configuration=strategy_config)
    except ValidationError as e:
        raise FidesopsValidationError(message=str(e))


def get_strategy_names() -> List[str]:
    """Returns all supported authentication strategies"""
    return [s.name for s in SupportedAuthenticationStrategies]



from typing import Optional

from loguru import logger
from pydantic import ValidationError
from sqlalchemy import Column, Enum, String
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.mutable import MutableDict
from sqlalchemy.orm import Session
from sqlalchemy_utils.types.encrypted.encrypted_type import (
    AesGcmEngine,
    StringEncryptedType,
)

from fides.api.ops.common_exceptions import MessageDispatchException
from fides.api.ops.db.base_class import JSONTypeOverride
from fides.api.ops.schemas.messaging.messaging import (
    EMAIL_MESSAGING_SERVICES,
    SMS_MESSAGING_SERVICES,
    SUPPORTED_MESSAGING_SERVICE_SECRETS,
    MessagingMethod,
    MessagingServiceSecretsMailgun,
    MessagingServiceSecretsTwilioEmail,
    MessagingServiceSecretsTwilioSMS,
    MessagingServiceType,
)
from fides.api.ops.schemas.messaging.messaging_secrets_docs_only import (
    possible_messaging_secrets,
)
from fides.api.ops.util.logger import Pii
from fides.core.config import get_config
from fides.lib.db.base import Base  # type: ignore[attr-defined]

CONFIG = get_config()


def get_messaging_method(
    service_type: Optional[str],
) -> Optional[MessagingMethod]:
    """returns messaging method based on configured service type"""
    if service_type in EMAIL_MESSAGING_SERVICES:
        return MessagingMethod.EMAIL
    if service_type in SMS_MESSAGING_SERVICES:
        return MessagingMethod.SMS
    return None


def get_schema_for_secrets(
    service_type: MessagingServiceType,
    secrets: possible_messaging_secrets,
) -> SUPPORTED_MESSAGING_SERVICE_SECRETS:
    """
    Returns the secrets that pertain to `service_type` represented as a Pydantic schema
    for validation purposes.
    """
    try:
        schema = {
            MessagingServiceType.MAILGUN: MessagingServiceSecretsMailgun,
            MessagingServiceType.TWILIO_TEXT: MessagingServiceSecretsTwilioSMS,
            MessagingServiceType.TWILIO_EMAIL: MessagingServiceSecretsTwilioEmail,
        }[service_type]
    except KeyError:
        raise ValueError(
            f"`service_type` {service_type} has no supported `secrets` validation."
        )

    try:
        return schema.parse_obj(secrets)  # type: ignore
    except ValidationError as exc:
        # Pydantic requires validators raise either a ValueError, TypeError, or AssertionError
        # so this exception is cast into a `ValueError`.
        errors = [f"{err['msg']} {str(err['loc'])}" for err in exc.errors()]
        raise ValueError(errors)


class MessagingConfig(Base):
    """The DB ORM model for MessagingConfig"""

    key = Column(String, index=True, unique=True, nullable=False)
    name = Column(String, unique=True, index=True)
    service_type = Column(
        Enum(MessagingServiceType), index=True, unique=True, nullable=False
    )
    details = Column(MutableDict.as_mutable(JSONB), nullable=True)
    secrets = Column(
        MutableDict.as_mutable(
            StringEncryptedType(
                JSONTypeOverride,
                CONFIG.security.app_encryption_key,
                AesGcmEngine,
                "pkcs5",
            )
        ),
        nullable=True,
    )  # Type bytea in the db

    @classmethod
    def get_configuration(cls, db: Session, service_type: str) -> Base:
        """
        Fetches the configured MessagingConfig record by service type. Once fetched this function validates that
        the MessagingConfig is configured with secrets.
        """
        instance: Optional[Base] = cls.get_by(
            db=db, field="service_type", value=service_type
        )
        if not instance:
            raise MessageDispatchException(
                f"No messaging config found for service_type {service_type}."
            )
        if not instance.secrets:
            logger.warning(
                "Messaging secrets not found for config with key: {}", instance.key
            )
            raise MessageDispatchException(
                f"Messaging secrets not found for config with key: {instance.key}"
            )
        return instance

    def set_secrets(
        self,
        *,
        db: Session,
        messaging_secrets: possible_messaging_secrets,
    ) -> None:
        """Creates or updates secrets associated with a config id"""

        service_type = self.service_type
        if not service_type:
            raise ValueError(
                "This object must have a `service_type` to validate secrets."
            )

        try:
            get_schema_for_secrets(
                service_type=service_type,  # type: ignore
                secrets=messaging_secrets,
            )
        except (
            KeyError,
            ValidationError,
        ) as exc:
            logger.error("Error: {}", Pii(str(exc)))
            # We don't want to handle these explicitly here, only in the API view
            raise

        self.secrets = messaging_secrets
        self.save(db=db)



"""Add consent

Revision ID: a0e6feb5bdc8
Revises: 7a4f4042091e
Create Date: 2022-09-12 14:46:05.443579

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "a0e6feb5bdc8"
down_revision = "7a4f4042091e"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "consent",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("provided_identity_id", sa.String(), nullable=True),
        sa.Column("data_use", sa.String(), nullable=False),
        sa.Column("data_use_description", sa.String(), nullable=True),
        sa.Column("opt_in", sa.Boolean(), nullable=False),
        sa.ForeignKeyConstraint(
            ["provided_identity_id"],
            ["providedidentity.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("data_use"),
    )
    op.create_index(op.f("ix_consent_id"), "consent", ["id"], unique=False)
    op.alter_column(
        "providedidentity",
        "privacy_request_id",
        existing_type=sa.VARCHAR(),
        nullable=True,
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column(
        "providedidentity",
        "privacy_request_id",
        existing_type=sa.VARCHAR(),
        nullable=False,
    )
    op.drop_index(op.f("ix_consent_id"), table_name="consent")
    op.drop_table("consent")
    # ### end Alembic commands ###



from typing import Any, Dict

from fastapi.params import Security
from loguru import logger

from fides.api.ops.api.v1 import scope_registry as scopes
from fides.api.ops.api.v1 import urn_registry as urls
from fides.api.ops.util.api_router import APIRouter
from fides.api.ops.util.oauth_util import verify_oauth_client
from fides.core.config import censor_config
from fides.core.config import get_config as get_app_config

router = APIRouter(tags=["Config"], prefix=urls.V1_URL_PREFIX)


@router.get(
    urls.CONFIG,
    dependencies=[Security(verify_oauth_client, scopes=[scopes.CONFIG_READ])],
    response_model=Dict[str, Any],
)
def get_config() -> Dict[str, Any]:
    """Returns the current API exposable Fides configuration."""
    logger.info("Getting the exposable Fides configuration")
    config = censor_config(get_app_config())
    return config



"""Add custom metadata tables

Revision ID: a08024ba7e2b
Revises: 3caf11127442
Create Date: 2023-01-18 16:10:23.832614

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "a08024ba7e2b"
down_revision = "3caf11127442"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "plus_custom_field_value_list",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("description", sa.String(), nullable=True),
        sa.Column("allowed_values", sa.ARRAY(sa.String()), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_plus_custom_field_value_list_id"),
        "plus_custom_field_value_list",
        ["id"],
        unique=False,
    )
    op.create_table(
        "plus_custom_field_definition",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("description", sa.String(), nullable=True),
        sa.Column(
            "field_type",
            sa.Enum("string", "string_list", name="allowedtypes"),
            nullable=False,
        ),
        sa.Column("allow_list_id", sa.String(), nullable=True),
        sa.Column(
            "resource_type",
            sa.Enum(
                "system",
                "data_use",
                "data_category",
                "data_subject",
                name="resourcetypes",
            ),
            nullable=False,
        ),
        sa.Column("field_definition", sa.String(), nullable=True),
        sa.Column("active", sa.BOOLEAN(), nullable=False),
        sa.ForeignKeyConstraint(
            ["allow_list_id"],
            ["plus_custom_field_value_list.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_plus_custom_field_definition_field_definition"),
        "plus_custom_field_definition",
        ["field_definition"],
        unique=False,
    )
    op.create_index(
        op.f("ix_plus_custom_field_definition_id"),
        "plus_custom_field_definition",
        ["id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_plus_custom_field_definition_name"),
        "plus_custom_field_definition",
        ["name"],
        unique=False,
    )
    op.create_table(
        "plus_custom_field",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "resource_type",
            sa.Enum(
                "system",
                "data_use",
                "data_category",
                "data_subject",
                name="resourcetypes",
            ),
            nullable=False,
        ),
        sa.Column("resource_id", sa.String(), nullable=False),
        sa.Column("custom_field_definition_id", sa.String(), nullable=False),
        sa.Column("value", sa.ARRAY(sa.String()), nullable=True),
        sa.ForeignKeyConstraint(
            ["custom_field_definition_id"],
            ["plus_custom_field_definition.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_plus_custom_field_id"), "plus_custom_field", ["id"], unique=False
    )
    op.create_index(
        op.f("ix_plus_custom_field_resource_id"),
        "plus_custom_field",
        ["resource_id"],
        unique=False,
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(
        op.f("ix_plus_custom_field_resource_id"), table_name="plus_custom_field"
    )
    op.drop_index(op.f("ix_plus_custom_field_id"), table_name="plus_custom_field")
    op.drop_table("plus_custom_field")
    op.drop_index(
        op.f("ix_plus_custom_field_definition_name"),
        table_name="plus_custom_field_definition",
    )
    op.drop_index(
        op.f("ix_plus_custom_field_definition_id"),
        table_name="plus_custom_field_definition",
    )
    op.drop_index(
        op.f("ix_plus_custom_field_definition_field_definition"),
        table_name="plus_custom_field_definition",
    )
    op.drop_table("plus_custom_field_definition")
    op.drop_index(
        op.f("ix_plus_custom_field_value_list_id"),
        table_name="plus_custom_field_value_list",
    )
    op.drop_table("plus_custom_field_value_list")
    # ### end Alembic commands ###



"""Contains all of the CRUD-type CLI commands for fides."""
import click
import yaml

from fides.cli.options import fides_key_argument, resource_type_argument
from fides.cli.utils import handle_cli_response, print_divider, with_analytics
from fides.core import api as _api
from fides.core.api_helpers import get_server_resource, list_server_resources
from fides.core.utils import echo_green


@click.command()
@click.pass_context
@resource_type_argument
@fides_key_argument
@with_analytics
def delete(ctx: click.Context, resource_type: str, fides_key: str) -> None:
    """
    Delete a resource on the server.
    """
    config = ctx.obj["CONFIG"]
    handle_cli_response(
        _api.delete(
            url=config.cli.server_url,
            resource_type=resource_type,
            resource_id=fides_key,
            headers=config.user.auth_header,
        )
    )


@click.command(name="get")
@click.pass_context
@resource_type_argument
@fides_key_argument
@with_analytics
def get_resource(ctx: click.Context, resource_type: str, fides_key: str) -> None:
    """
    View a resource from the server as a YAML object.
    """
    config = ctx.obj["CONFIG"]
    resource = get_server_resource(
        url=config.cli.server_url,
        resource_type=resource_type,
        resource_key=fides_key,
        headers=config.user.auth_header,
        raw=True,
    )
    print_divider()
    echo_green(yaml.dump({resource_type: [resource]}))


@click.command(name="ls")
@click.pass_context
@resource_type_argument
@with_analytics
def list_resources(ctx: click.Context, resource_type: str) -> None:
    """
    Get a list of all resources of this type from the server and display them as YAML.
    """
    config = ctx.obj["CONFIG"]
    resources = list_server_resources(
        url=config.cli.server_url,
        resource_type=resource_type,
        headers=config.user.auth_header,
        exclude_keys=[],
        raw=True,
    )
    print_divider()
    echo_green(yaml.dump({resource_type: resources}))



"""The root module for the Fides package."""
from fides.core.config import get_config

from ._version import get_versions

# Load the config here as a work around to the timing issue with environment variables
get_config()

__version__ = get_versions()["version"]
del get_versions



"""another upstream merge

Revision ID: 6f98e1bd7e4f
Revises: 6bd93cb0603d, 794e5f5b76ae
Create Date: 2022-09-22 20:46:34.831732

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "6f98e1bd7e4f"
down_revision = ("6bd93cb0603d", "794e5f5b76ae")
branch_labels = None
depends_on = None


def upgrade():
    pass


def downgrade():
    pass



from enum import Enum
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, root_validator, validator

from fides.api.ops.schemas.saas.saas_config import Header, QueryParam, SaaSRequest
from fides.api.ops.schemas.saas.shared_schemas import (
    ConnectorParamRef,
    IdentityParamRef,
)


class StrategyConfiguration(BaseModel):
    """Base class for strategy configuration"""


class UnwrapPostProcessorConfiguration(StrategyConfiguration):
    """Dynamic JSON path access"""

    data_path: str


class FilterPostProcessorConfiguration(StrategyConfiguration):
    """Returns objects where a field has a given value"""

    field: str
    value: Union[str, IdentityParamRef]
    exact: bool = True
    case_sensitive: bool = True


class OffsetPaginationConfiguration(StrategyConfiguration):
    """
    Increases the value of the query param `incremental_param` by the `increment_by` until the `limit` is hit
    or there is no more data available.
    """

    incremental_param: str
    increment_by: int
    limit: Optional[Union[int, ConnectorParamRef]]

    @validator("increment_by")
    def check_increment_by(cls, increment_by: int) -> int:
        if increment_by == 0:
            raise ValueError("'increment_by' cannot be zero")
        if increment_by < 0:
            raise ValueError("'increment_by' cannot be negative")
        return increment_by


class LinkSource(Enum):
    """Locations where the link to the next page may be found."""

    headers = "headers"
    body = "body"


class LinkPaginationConfiguration(StrategyConfiguration):
    """Gets the URL for the next page from the headers or the body."""

    source: LinkSource
    rel: Optional[str]
    path: Optional[str]

    @root_validator
    def validate_fields(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        source = values.get("source")
        if source == LinkSource.headers.value and values.get("rel") is None:
            raise ValueError(
                "The 'rel' value must be specified when accessing the link from the headers"
            )
        if source == LinkSource.body.value and values.get("path") is None:
            raise ValueError(
                "The 'path' value must be specified when accessing the link from the body"
            )
        return values

    class Config:
        """Using enum values"""

        use_enum_values = True


class CursorPaginationConfiguration(StrategyConfiguration):
    """
    Extracts the cursor value from the 'field' of the last object in the array specified by 'data_path'
    """

    cursor_param: str
    field: str


class ApiKeyAuthenticationConfiguration(StrategyConfiguration):
    """
    API key parameter to be added in as a header or query param
    """

    headers: Optional[List[Header]]
    query_params: Optional[List[QueryParam]]
    body: Optional[str]

    @root_validator
    def validate_fields(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        headers = values.get("headers")
        query_params = values.get("query_params")
        body = values.get("body")

        if not headers and not query_params and not body:
            raise ValueError(
                "At least one 'header', 'query_param', or 'body' object must be defined in an 'api_key' auth configuration"
            )

        return values


class BasicAuthenticationConfiguration(StrategyConfiguration):
    """
    Username and password to add basic authentication to HTTP requests
    """

    username: str
    password: Optional[str]


class BearerAuthenticationConfiguration(StrategyConfiguration):
    """
    Token to add as bearer authentication for HTTP requests
    """

    token: str


class QueryParamAuthenticationConfiguration(StrategyConfiguration):
    """
    Value to add as query param for HTTP requests
    """

    name: str
    value: str


class OAuth2BaseConfiguration(StrategyConfiguration):
    """
    OAuth2 endpoints for token retrieval, and token refresh.
    Includes an optional expires_in parameter (in seconds) for OAuth2 integrations that
    do not specify a TTL for the access tokens.
    """

    expires_in: Optional[int]
    token_request: SaaSRequest
    refresh_request: Optional[SaaSRequest]


class OAuth2AuthorizationCodeConfiguration(OAuth2BaseConfiguration):
    """
    The standard OAuth2 configuration but with an additional property to configure
    the authorization request for the Authorization Code flow.
    """

    authorization_request: SaaSRequest



from typing import Any, Dict

from boto3 import Session
from loguru import logger

from fides.api.ops.common_exceptions import StorageUploadError
from fides.api.ops.schemas.storage.storage import S3AuthMethod, StorageSecrets


def get_s3_session(
    auth_method: S3AuthMethod, storage_secrets: Dict[StorageSecrets, Any]
) -> Session:
    """Abstraction to retrieve s3 session using secrets"""
    if auth_method == S3AuthMethod.SECRET_KEYS.value:

        if storage_secrets is None:
            err_msg = "Storage secrets not found for S3 storage."
            logger.warning(err_msg)
            raise StorageUploadError(err_msg)

        session = Session(
            aws_access_key_id=storage_secrets[StorageSecrets.AWS_ACCESS_KEY_ID.value],  # type: ignore
            aws_secret_access_key=storage_secrets[
                StorageSecrets.AWS_SECRET_ACCESS_KEY.value  # type: ignore
            ],
        )

        # Check that credentials are valid
        client = session.client("sts")
        client.get_caller_identity()
        return session

    if auth_method == S3AuthMethod.AUTOMATIC.value:
        session = Session()
        logger.info("Successfully created automatic session")
        return session

    logger.error("Auth method not supported for S3: {}", auth_method)
    raise ValueError(f"Auth method not supported for S3: {auth_method}")



# pylint: disable=useless-import-alias

# Because these modules are imported into the __init__.py and used elsewhere they need
# to be explicitly exported in order to prevent implicit reexport errors in mypy.
# This is done by importing "as": `from fides.module import MyClass as MyClass`.

from typing import Any, Dict, Optional, Union

from fides.api.ops.models.connectionconfig import ConnectionType
from fides.api.ops.schemas.connection_configuration.connection_secrets import (
    ConnectionConfigSecretsSchema as ConnectionConfigSecretsSchema,
)
from fides.api.ops.schemas.connection_configuration.connection_secrets_bigquery import (
    BigQueryDocsSchema as BigQueryDocsSchema,
)
from fides.api.ops.schemas.connection_configuration.connection_secrets_bigquery import (
    BigQuerySchema as BigQuerySchema,
)
from fides.api.ops.schemas.connection_configuration.connection_secrets_email import (
    EmailDocsSchema as EmailDocsSchema,
)
from fides.api.ops.schemas.connection_configuration.connection_secrets_email import (
    EmailSchema as EmailSchema,
)
from fides.api.ops.schemas.connection_configuration.connection_secrets_fides import (
    FidesConnectorSchema,
    FidesDocsSchema,
)
from fides.api.ops.schemas.connection_configuration.connection_secrets_manual_webhook import (
    ManualWebhookSchema as ManualWebhookSchema,
)
from fides.api.ops.schemas.connection_configuration.connection_secrets_manual_webhook import (
    ManualWebhookSchemaforDocs as ManualWebhookSchemaforDocs,
)
from fides.api.ops.schemas.connection_configuration.connection_secrets_mariadb import (
    MariaDBDocsSchema as MariaDBDocsSchema,
)
from fides.api.ops.schemas.connection_configuration.connection_secrets_mariadb import (
    MariaDBSchema as MariaDBSchema,
)
from fides.api.ops.schemas.connection_configuration.connection_secrets_mongodb import (
    MongoDBDocsSchema as MongoDBDocsSchema,
)
from fides.api.ops.schemas.connection_configuration.connection_secrets_mongodb import (
    MongoDBSchema as MongoDBSchema,
)
from fides.api.ops.schemas.connection_configuration.connection_secrets_mssql import (
    MicrosoftSQLServerSchema as MicrosoftSQLServerSchema,
)
from fides.api.ops.schemas.connection_configuration.connection_secrets_mssql import (
    MSSQLDocsSchema as MSSQLDocsSchema,
)
from fides.api.ops.schemas.connection_configuration.connection_secrets_mysql import (
    MySQLDocsSchema as MySQLDocsSchema,
)
from fides.api.ops.schemas.connection_configuration.connection_secrets_mysql import (
    MySQLSchema as MySQLSchema,
)
from fides.api.ops.schemas.connection_configuration.connection_secrets_postgres import (
    PostgreSQLDocsSchema as PostgreSQLDocsSchema,
)
from fides.api.ops.schemas.connection_configuration.connection_secrets_postgres import (
    PostgreSQLSchema as PostgreSQLSchema,
)
from fides.api.ops.schemas.connection_configuration.connection_secrets_redshift import (
    RedshiftDocsSchema as RedshiftDocsSchema,
)
from fides.api.ops.schemas.connection_configuration.connection_secrets_redshift import (
    RedshiftSchema as RedshiftSchema,
)
from fides.api.ops.schemas.connection_configuration.connection_secrets_saas import (
    SaaSSchema as SaaSSchema,
)
from fides.api.ops.schemas.connection_configuration.connection_secrets_saas import (
    SaaSSchemaFactory as SaaSSchemaFactory,
)
from fides.api.ops.schemas.connection_configuration.connection_secrets_snowflake import (
    SnowflakeDocsSchema as SnowflakeDocsSchema,
)
from fides.api.ops.schemas.connection_configuration.connection_secrets_snowflake import (
    SnowflakeSchema as SnowflakeSchema,
)
from fides.api.ops.schemas.connection_configuration.connection_secrets_timescale import (
    TimescaleDocsSchema as TimescaleDocsSchema,
)
from fides.api.ops.schemas.connection_configuration.connection_secrets_timescale import (
    TimescaleSchema as TimescaleSchema,
)
from fides.api.ops.schemas.connection_configuration.connections_secrets_https import (
    HttpsSchema as HttpsSchema,
)
from fides.api.ops.schemas.saas.saas_config import SaaSConfig as SaaSConfig

secrets_schemas: Dict[str, Any] = {
    ConnectionType.postgres.value: PostgreSQLSchema,
    ConnectionType.https.value: HttpsSchema,
    ConnectionType.mongodb.value: MongoDBSchema,
    ConnectionType.mysql.value: MySQLSchema,
    ConnectionType.redshift.value: RedshiftSchema,
    ConnectionType.snowflake.value: SnowflakeSchema,
    ConnectionType.mssql.value: MicrosoftSQLServerSchema,
    ConnectionType.mariadb.value: MariaDBSchema,
    ConnectionType.bigquery.value: BigQuerySchema,
    ConnectionType.saas.value: SaaSSchema,
    ConnectionType.email.value: EmailSchema,
    ConnectionType.manual_webhook.value: ManualWebhookSchema,
    ConnectionType.timescale.value: TimescaleSchema,
    ConnectionType.fides.value: FidesConnectorSchema,
}


def get_connection_secrets_schema(
    connection_type: str, saas_config: Optional[SaaSConfig]
) -> ConnectionConfigSecretsSchema:
    """
    Returns a validation schema for the connection "secrets" depending on the connection_type.

    For example, a "postgres" connection_type would need to have its secrets
    validated against a PostgreSQL schema.
    """
    if connection_type == "saas" and not saas_config:
        raise ValueError(
            "A SaaS config to validate the secrets is required for a saas connection"
        )
    try:
        schema = (
            SaaSSchemaFactory(saas_config).get_saas_schema()
            if saas_config
            else secrets_schemas[connection_type]
        )
        return schema
    except KeyError:
        raise NotImplementedError(
            f"Add {connection_type} to the 'secrets_schema' mapping."
        )


connection_secrets_schemas = Union[
    MongoDBDocsSchema,
    PostgreSQLDocsSchema,
    MySQLDocsSchema,
    RedshiftDocsSchema,
    SnowflakeDocsSchema,
    MSSQLDocsSchema,
    MariaDBDocsSchema,
    BigQueryDocsSchema,
    SaaSSchema,
    EmailDocsSchema,
    ManualWebhookSchemaforDocs,
    TimescaleDocsSchema,
    FidesDocsSchema,
]



"""update dataset field name

Revision ID: 732105cd54e3
Revises: e576b6a80a49
Create Date: 2021-10-01 06:48:22.449643

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "732105cd54e3"
down_revision = "e576b6a80a49"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "datasets", sa.Column("data_qualifiers", sa.ARRAY(sa.String()), nullable=True)
    )
    op.drop_column("datasets", "data_qualifier")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "datasets",
        sa.Column(
            "data_qualifier",
            postgresql.ARRAY(sa.VARCHAR()),
            autoincrement=False,
            nullable=True,
        ),
    )
    op.drop_column("datasets", "data_qualifiers")
    # ### end Alembic commands ###



import base64
import pickle
from typing import Any, Dict, List, Optional, Union

from loguru import logger
from redis import Redis
from redis.client import Script  # type: ignore
from redis.exceptions import ConnectionError as ConnectionErrorFromRedis

from fides.api.ops import common_exceptions
from fides.api.ops.schemas.masking.masking_secrets import SecretType
from fides.core.config import get_config

CONFIG = get_config()

# This constant represents every type a redis key may contain, and can be
# extended if needed
RedisValue = Union[bytes, float, int, str]

_connection = None


class FidesopsRedis(Redis):
    """
    An extension to Redis' python bindings to support auto expiring data input. This class
    should never be instantiated on its own.
    """

    def set_with_autoexpire(
        self,
        key: str,
        value: RedisValue,
        expire_time: int = CONFIG.redis.default_ttl_seconds,
    ) -> Optional[bool]:
        """Call the connection class' default set method with ex= our default TTL"""
        if not expire_time:
            # We have to check this condition for the edge case where `None` is explicitly
            # passed to this method.
            expire_time = CONFIG.redis.default_ttl_seconds
        return self.set(key, value, ex=expire_time)

    def get_keys_by_prefix(self, prefix: str, chunk_size: int = 1000) -> List[str]:
        """Retrieve all keys that match a given prefix."""
        cursor: Any = "0"
        out = []
        while cursor != 0:
            cursor, keys = self.scan(
                cursor=cursor, match=f"{prefix}*", count=chunk_size
            )
            out.extend(keys)
        return out

    def delete_keys_by_prefix(self, prefix: str) -> None:
        """Delete all keys starting with a given prefix"""
        s: Script = self.register_script(
            f"for _,k in ipairs(redis.call('keys','{prefix}*')) do redis.call('del',k) end"
        )
        s()

    def get_values(self, keys: List[str]) -> Dict[str, Optional[Any]]:
        """Retrieve all values corresponding to the set of input keys and return them as a
        dictionary. Note that if a key does not exist in redis it will be returned as None"""
        values = self.mget(keys)
        return {x[0]: x[1] for x in zip(keys, values)}

    def set_encoded_object(self, key: str, obj: Any) -> Optional[bool]:
        """Set an object in redis in an encoded form. This object should be retrieved via
        get_objects_by_prefix or processed with decode_obj."""
        return self.set_with_autoexpire(f"EN_{key}", FidesopsRedis.encode_obj(obj))

    def get_encoded_by_key(self, key: str) -> Optional[Any]:
        """Returns cached obj decoded from base64"""
        val = super().get(key)
        return self.decode_obj(val) if val else None

    def get_encoded_objects_by_prefix(self, prefix: str) -> Dict[str, Optional[Any]]:
        """Return all objects stored under a given prefix. This method
        assumes these objects have been stored encoded using set_object"""
        keys = self.get_keys_by_prefix(f"EN_{prefix}")
        encoded_object_dict = self.get_values(keys)
        return {
            key: FidesopsRedis.decode_obj(value)
            for key, value in encoded_object_dict.items()
        }

    @staticmethod
    def encode_obj(obj: Any) -> bytes:
        """Encode an object to a base64 string that can be stored in Redis"""
        return base64.b64encode(pickle.dumps(obj))

    @staticmethod
    def decode_obj(bs: Optional[bytes]) -> Any:
        """Decode an object from its base64 representation.

        Since Redis may not contain a value
        for a given key it's possible we may try to decode an empty object."""
        if bs:
            return pickle.loads(base64.b64decode(bs))
        return None


def get_cache() -> FidesopsRedis:
    """Return a singleton connection to our Redis cache"""
    global _connection  # pylint: disable=W0603
    if _connection is None:
        logger.debug("Creating new Redis connection...")
        _connection = FidesopsRedis(  # type: ignore[call-overload]
            charset=CONFIG.redis.charset,
            decode_responses=CONFIG.redis.decode_responses,
            host=CONFIG.redis.host,
            port=CONFIG.redis.port,
            db=CONFIG.redis.db_index,
            password=CONFIG.redis.password,
            ssl=CONFIG.redis.ssl,
            ssl_cert_reqs=CONFIG.redis.ssl_cert_reqs,
        )
        logger.debug("New Redis connection created.")

    logger.debug("Testing Redis connection...")
    try:
        connected = _connection.ping()
    except ConnectionErrorFromRedis:
        connected = False
    else:
        logger.debug("Redis connection succeeded.")

    if not connected:
        logger.debug("Redis connection failed.")
        raise common_exceptions.RedisConnectionError(
            "Unable to establish Redis connection. Fidesops is unable to accept PrivacyRequsts."
        )

    return _connection


def get_identity_cache_key(privacy_request_id: str, identity_attribute: str) -> str:
    """Return the key at which to save this PrivacyRequest's identity for the passed in attribute"""
    # TODO: Remove this prefix
    return f"id-{privacy_request_id}-identity-{identity_attribute}"


def get_drp_request_body_cache_key(
    privacy_request_id: str, identity_attribute: str
) -> str:
    """Return the key at which to save this PrivacyRequest's drp request body for the passed in attribute"""
    return f"id-{privacy_request_id}-drp-{identity_attribute}"


def get_encryption_cache_key(privacy_request_id: str, encryption_attr: str) -> str:
    """Return the key at which to save this PrivacyRequest's encryption attribute"""
    return f"id-{privacy_request_id}-encryption-{encryption_attr}"


def get_masking_secret_cache_key(
    privacy_request_id: str, masking_strategy: str, secret_type: SecretType
) -> str:
    """Return the key at which to save this PrivacyRequest's masking secret attribute"""
    return (
        f"id-{privacy_request_id}-masking-secret-{masking_strategy}-{secret_type.value}"
    )


def get_all_cache_keys_for_privacy_request(privacy_request_id: str) -> List[Any]:
    """Returns all cache keys related to this privacy request's cached identities"""
    cache: FidesopsRedis = get_cache()
    return cache.keys(f"{privacy_request_id}-*") + cache.keys(
        f"id-{privacy_request_id}-*"
    )


def get_async_task_tracking_cache_key(privacy_request_id: str) -> str:
    return f"id-{privacy_request_id}-async-execution"



from requests import PreparedRequest

from fides.api.ops.models.connectionconfig import ConnectionConfig
from fides.api.ops.schemas.saas.strategy_configuration import OAuth2BaseConfiguration
from fides.api.ops.service.authentication.authentication_strategy_oauth2_base import (
    OAuth2AuthenticationStrategyBase,
)


class OAuth2ClientCredentialsAuthenticationStrategy(OAuth2AuthenticationStrategyBase):
    """
    Checks the expiration date on the stored access token and refreshes
    it if needed using the configured token refresh request.
    """

    name = "oauth2_client_credentials"
    configuration_model = OAuth2BaseConfiguration

    def add_authentication(
        self, request: PreparedRequest, connection_config: ConnectionConfig
    ) -> PreparedRequest:
        """
        Checks the expiration date on the existing access token and refreshes if necessary.
        The existing/updated access token is then added to the request as a bearer token.
        """

        access_token = connection_config.secrets.get("access_token")  # type: ignore
        if not access_token:
            access_token = self.get_access_token(connection_config)
        else:
            access_token = self._refresh_token(connection_config)

        # add access_token to request
        request.headers["Authorization"] = "Bearer " + access_token
        return request



import random
from datetime import datetime, timedelta
from typing import Any, Dict, List, Optional, Set, Tuple

import requests
from loguru import logger
from pydantic import ValidationError
from redis.exceptions import DataError
from sqlalchemy.orm import Session

from fides.api.ops import common_exceptions
from fides.api.ops.api.v1.urn_registry import (
    PRIVACY_REQUEST_TRANSFER_TO_PARENT,
    V1_URL_PREFIX,
)
from fides.api.ops.common_exceptions import (
    ClientUnsuccessfulException,
    IdentityNotFoundException,
    ManualWebhookFieldsUnset,
    MessageDispatchException,
    NoCachedManualWebhookEntry,
    PrivacyRequestPaused,
)
from fides.api.ops.graph.analytics_events import (
    failed_graph_analytics_event,
    fideslog_graph_failure,
)
from fides.api.ops.graph.config import CollectionAddress, GraphDataset
from fides.api.ops.graph.graph import DatasetGraph
from fides.api.ops.models.connectionconfig import ConnectionConfig, ConnectionType
from fides.api.ops.models.datasetconfig import DatasetConfig
from fides.api.ops.models.manual_webhook import AccessManualWebhook
from fides.api.ops.models.policy import (
    ActionType,
    CurrentStep,
    Policy,
    PolicyPostWebhook,
    PolicyPreWebhook,
    WebhookTypes,
)
from fides.api.ops.models.privacy_request import (
    PrivacyRequest,
    PrivacyRequestStatus,
    ProvidedIdentityType,
    can_run_checkpoint,
)
from fides.api.ops.schemas.messaging.messaging import (
    AccessRequestCompleteBodyParams,
    MessagingActionType,
)
from fides.api.ops.schemas.redis_cache import Identity
from fides.api.ops.service.connectors import FidesConnector
from fides.api.ops.service.connectors.email_connector import (
    email_connector_erasure_send,
)
from fides.api.ops.service.connectors.fides_connector import (
    filter_fides_connector_datasets,
)
from fides.api.ops.service.messaging.message_dispatch_service import dispatch_message
from fides.api.ops.service.storage.storage_uploader_service import upload
from fides.api.ops.task.filter_results import filter_data_categories
from fides.api.ops.task.graph_task import (
    get_cached_data_for_erasures,
    run_access_request,
    run_consent_request,
    run_erasure,
)
from fides.api.ops.tasks import DatabaseTask, celery_app
from fides.api.ops.tasks.scheduled.scheduler import scheduler
from fides.api.ops.util.cache import (
    FidesopsRedis,
    get_async_task_tracking_cache_key,
    get_cache,
)
from fides.api.ops.util.collection_util import Row
from fides.api.ops.util.logger import Pii, _log_exception, _log_warning
from fides.api.ops.util.wrappers import sync
from fides.core.config import get_config
from fides.lib.db.session import get_db_session
from fides.lib.models.audit_log import AuditLog, AuditLogAction
from fides.lib.schemas.base_class import BaseSchema

CONFIG = get_config()


class ManualWebhookResults(BaseSchema):
    """Represents manual webhook data retrieved from the cache and whether privacy request execution should continue"""

    manual_data: Dict[str, List[Dict[str, Optional[Any]]]]
    proceed: bool


def get_access_manual_webhook_inputs(
    db: Session, privacy_request: PrivacyRequest, policy: Policy
) -> ManualWebhookResults:

    """Retrieves manually uploaded data for all AccessManualWebhooks and formats in a way
    to match automatically retrieved data (a list of rows). Also returns if execution should proceed.

    This data will be uploaded to the user as-is, without filtering.
    """
    manual_inputs: Dict[str, List[Dict[str, Optional[Any]]]] = {}

    if not policy.get_rules_for_action(action_type=ActionType.access):
        # Don't fetch manual inputs unless this policy has an access rule
        return ManualWebhookResults(manual_data=manual_inputs, proceed=True)

    try:
        for manual_webhook in AccessManualWebhook.get_enabled(db):
            manual_inputs[manual_webhook.connection_config.key] = [
                privacy_request.get_manual_webhook_input_strict(manual_webhook)
            ]
    except (
        NoCachedManualWebhookEntry,
        ValidationError,
        ManualWebhookFieldsUnset,
    ) as exc:
        logger.info(exc)
        privacy_request.status = PrivacyRequestStatus.requires_input
        privacy_request.save(db)
        return ManualWebhookResults(manual_data=manual_inputs, proceed=False)

    return ManualWebhookResults(manual_data=manual_inputs, proceed=True)


def run_webhooks_and_report_status(
    db: Session,
    privacy_request: PrivacyRequest,
    webhook_cls: WebhookTypes,
    after_webhook_id: Optional[str] = None,
) -> bool:
    """
    Runs a series of webhooks either pre- or post- privacy request execution, if any are configured.
    Updates privacy request status if execution is paused/errored.
    Returns True if execution should proceed.
    """
    webhooks = db.query(webhook_cls).filter_by(policy_id=privacy_request.policy.id)  # type: ignore

    if after_webhook_id:
        # Only run webhooks configured to run after this Pre-Execution webhook
        pre_webhook = PolicyPreWebhook.get(db=db, object_id=after_webhook_id)
        webhooks = webhooks.filter(  # type: ignore[call-arg]
            webhook_cls.order > pre_webhook.order,  # type: ignore[union-attr]
        )

    current_step = CurrentStep[f"{webhook_cls.prefix}_webhooks"]

    for webhook in webhooks.order_by(webhook_cls.order):  # type: ignore[union-attr]
        try:
            privacy_request.trigger_policy_webhook(webhook)
        except PrivacyRequestPaused:
            logger.info(
                "Pausing execution of privacy request {}. Halt instruction received from webhook {}.",
                privacy_request.id,
                webhook.key,
            )
            privacy_request.pause_processing(db)
            initiate_paused_privacy_request_followup(privacy_request)
            return False
        except ClientUnsuccessfulException as exc:
            logger.error(
                "Privacy Request '{}' exited after response from webhook '{}': {}.",
                privacy_request.id,
                webhook.key,
                Pii(str(exc.args[0])),
            )
            privacy_request.error_processing(db)
            privacy_request.cache_failed_checkpoint_details(current_step)
            return False
        except ValidationError:
            logger.error(
                "Privacy Request '{}' errored due to response validation error from webhook '{}'.",
                privacy_request.id,
                webhook.key,
            )
            privacy_request.error_processing(db)
            privacy_request.cache_failed_checkpoint_details(current_step)
            return False

    return True


def upload_access_results(  # pylint: disable=R0912
    session: Session,
    policy: Policy,
    access_result: Dict[str, List[Row]],
    dataset_graph: DatasetGraph,
    privacy_request: PrivacyRequest,
    manual_data: Dict[str, List[Dict[str, Optional[Any]]]],
    fides_connector_datasets: Set[str],
) -> List[str]:
    """Process the data uploads after the access portion of the privacy request has completed"""
    download_urls: List[str] = []
    if not access_result:
        logger.info("No results returned for access request {}", privacy_request.id)

    for rule in policy.get_rules_for_action(  # pylint: disable=R1702
        action_type=ActionType.access
    ):
        if not rule.storage_destination:
            raise common_exceptions.RuleValidationError(
                f"No storage destination configured on rule {rule.key}"
            )
        target_categories: Set[str] = {target.data_category for target in rule.targets}  # type: ignore[attr-defined]
        filtered_results: Dict[
            str, List[Dict[str, Optional[Any]]]
        ] = filter_data_categories(
            access_result,
            target_categories,
            dataset_graph.data_category_field_mapping,
            rule.key,
            fides_connector_datasets,
        )

        filtered_results.update(
            manual_data
        )  # Add manual data directly to each upload packet

        logger.info(
            "Starting access request upload for rule {} for privacy request {}",
            rule.key,
            privacy_request.id,
        )
        try:
            download_url: Optional[str] = upload(
                db=session,
                request_id=privacy_request.id,
                data=filtered_results,
                storage_key=rule.storage_destination.key,  # type: ignore
            )
            if download_url:
                download_urls.append(download_url)
        except common_exceptions.StorageUploadError as exc:
            logger.error(
                "Error uploading subject access data for rule {} on policy {} and privacy request {} : {}",
                rule.key,
                policy.key,
                privacy_request.id,
                Pii(str(exc)),
            )
            privacy_request.status = PrivacyRequestStatus.error

    return download_urls


def queue_privacy_request(
    privacy_request_id: str,
    from_webhook_id: Optional[str] = None,
    from_step: Optional[str] = None,
) -> str:
    cache: FidesopsRedis = get_cache()
    logger.info("queueing privacy request")
    task = run_privacy_request.delay(
        privacy_request_id=privacy_request_id,
        from_webhook_id=from_webhook_id,
        from_step=from_step,
    )
    try:
        cache.set(
            get_async_task_tracking_cache_key(privacy_request_id),
            task.task_id,
        )
    except DataError:
        logger.debug(
            "Error tracking task_id for request with id {}", privacy_request_id
        )

    return task.task_id


@celery_app.task(base=DatabaseTask, bind=True)
@sync
async def run_privacy_request(
    self: DatabaseTask,
    privacy_request_id: str,
    from_webhook_id: Optional[str] = None,
    from_step: Optional[str] = None,
) -> None:
    # pylint: disable=too-many-statements, too-many-return-statements, too-many-branches
    """
    Dispatch a privacy_request into the execution layer by:
        1. Generate a graph from all the currently configured datasets
        2. Take the provided identity data
        3. Start the access request / erasure request execution
        4. When finished, upload the results to the configured storage destination if applicable

    Celery does not like for the function to be async so the @sync decorator runs the
    coroutine for it.
    """
    resume_step: Optional[CurrentStep] = CurrentStep(from_step) if from_step else None  # type: ignore
    if from_step:
        logger.info("Resuming privacy request from checkpoint: '{}'", from_step)

    with self.get_new_session() as session:
        privacy_request = PrivacyRequest.get(db=session, object_id=privacy_request_id)

        privacy_request.cache_failed_checkpoint_details()  # Reset failed step and collection to None

        if privacy_request.status == PrivacyRequestStatus.canceled:
            logger.info(
                "Terminating privacy request {}: request canceled.", privacy_request.id
            )
            return
        logger.info("Dispatching privacy request {}", privacy_request.id)
        privacy_request.start_processing(session)

        policy = privacy_request.policy
        manual_webhook_results: ManualWebhookResults = get_access_manual_webhook_inputs(
            session, privacy_request, policy
        )
        if not manual_webhook_results.proceed:
            return

        if can_run_checkpoint(
            request_checkpoint=CurrentStep.pre_webhooks, from_checkpoint=resume_step
        ):
            # Run pre-execution webhooks
            proceed = run_webhooks_and_report_status(
                session,
                privacy_request=privacy_request,
                webhook_cls=PolicyPreWebhook,  # type: ignore
                after_webhook_id=from_webhook_id,
            )
            if not proceed:
                return
        try:
            policy.rules[0]
        except IndexError:
            raise common_exceptions.MisconfiguredPolicyException(
                f"Policy with key {policy.key} must contain at least one Rule."
            )

        try:
            datasets = DatasetConfig.all(db=session)
            dataset_graphs = [dataset_config.get_graph() for dataset_config in datasets]
            dataset_graph = DatasetGraph(*dataset_graphs)
            identity_data = privacy_request.get_cached_identity_data()
            connection_configs = ConnectionConfig.all(db=session)
            fides_connector_datasets: Set[str] = filter_fides_connector_datasets(
                connection_configs
            )
            access_result_urls: List[str] = []

            if (
                policy.get_rules_for_action(action_type=ActionType.access)
                or policy.get_rules_for_action(action_type=ActionType.erasure)
            ) and can_run_checkpoint(
                request_checkpoint=CurrentStep.access, from_checkpoint=resume_step
            ):
                access_result: Dict[str, List[Row]] = await run_access_request(
                    privacy_request=privacy_request,
                    policy=policy,
                    graph=dataset_graph,
                    connection_configs=connection_configs,
                    identity=identity_data,
                    session=session,
                )
                access_result_urls = upload_access_results(
                    session,
                    policy,
                    access_result,
                    dataset_graph,
                    privacy_request,
                    manual_webhook_results.manual_data,
                    fides_connector_datasets,
                )

            if policy.get_rules_for_action(
                action_type=ActionType.erasure
            ) and can_run_checkpoint(
                request_checkpoint=CurrentStep.erasure, from_checkpoint=resume_step
            ):
                # We only need to run the erasure once until masking strategies are handled
                await run_erasure(
                    privacy_request=privacy_request,
                    policy=policy,
                    graph=dataset_graph,
                    connection_configs=connection_configs,
                    identity=identity_data,
                    access_request_data=get_cached_data_for_erasures(
                        privacy_request.id
                    ),
                    session=session,
                )

            if policy.get_rules_for_action(
                action_type=ActionType.consent
            ) and can_run_checkpoint(
                request_checkpoint=CurrentStep.consent,
                from_checkpoint=resume_step,
            ):
                await run_consent_request(
                    privacy_request=privacy_request,
                    policy=policy,
                    graph=build_consent_dataset_graph(datasets),
                    connection_configs=connection_configs,
                    identity=identity_data,
                    session=session,
                )

        except PrivacyRequestPaused as exc:
            privacy_request.pause_processing(session)
            _log_warning(exc, CONFIG.dev_mode)
            return

        except BaseException as exc:  # pylint: disable=broad-except
            privacy_request.error_processing(db=session)
            # Send analytics to Fideslog
            await fideslog_graph_failure(
                failed_graph_analytics_event(privacy_request, exc)
            )
            # If dev mode, log traceback
            _log_exception(exc, CONFIG.dev_mode)
            return

        # Send erasure requests via email to third parties where applicable
        if policy.get_rules_for_action(
            action_type=ActionType.erasure
        ) and can_run_checkpoint(
            request_checkpoint=CurrentStep.erasure_email_post_send,
            from_checkpoint=resume_step,
        ):
            try:
                email_connector_erasure_send(
                    db=session, privacy_request=privacy_request
                )
            except MessageDispatchException as exc:
                privacy_request.cache_failed_checkpoint_details(
                    step=CurrentStep.erasure_email_post_send, collection=None
                )
                privacy_request.error_processing(db=session)
                await fideslog_graph_failure(
                    failed_graph_analytics_event(privacy_request, exc)
                )
                # If dev mode, log traceback
                _log_exception(exc, CONFIG.dev_mode)
                return

        # Run post-execution webhooks
        proceed = run_webhooks_and_report_status(
            db=session,
            privacy_request=privacy_request,
            webhook_cls=PolicyPostWebhook,  # type: ignore
        )
        if not proceed:
            return

        if CONFIG.notifications.send_request_completion_notification:
            try:
                initiate_privacy_request_completion_email(
                    session, policy, access_result_urls, identity_data
                )
            except (IdentityNotFoundException, MessageDispatchException) as e:
                privacy_request.error_processing(db=session)
                # If dev mode, log traceback
                await fideslog_graph_failure(
                    failed_graph_analytics_event(privacy_request, e)
                )
                _log_exception(e, CONFIG.dev_mode)
                return
        privacy_request.finished_processing_at = datetime.utcnow()
        AuditLog.create(
            db=session,
            data={
                "user_id": "system",
                "privacy_request_id": privacy_request.id,
                "action": AuditLogAction.finished,
                "message": "",
            },
        )
        privacy_request.status = PrivacyRequestStatus.complete
        logger.info("Privacy request {} run completed.", privacy_request.id)
        privacy_request.save(db=session)


def build_consent_dataset_graph(datasets: List[DatasetConfig]) -> DatasetGraph:
    """
    Build the starting DatasetGraph for consent requests.

    Consent Graph has one node per dataset.  Nodes must be of saas type and have consent requests defined.
    """
    consent_datasets: List[GraphDataset] = []

    for dataset_config in datasets:
        connection_type: ConnectionType = (
            dataset_config.connection_config.connection_type  # type: ignore
        )
        saas_config: Optional[Dict] = dataset_config.connection_config.saas_config
        if (
            connection_type == ConnectionType.saas
            and saas_config
            and saas_config.get("consent_requests")
        ):
            consent_datasets.append(
                dataset_config.get_dataset_with_stubbed_collection()
            )

    return DatasetGraph(*consent_datasets)


def initiate_privacy_request_completion_email(
    session: Session,
    policy: Policy,
    access_result_urls: List[str],
    identity_data: Dict[str, Any],
) -> None:
    """
    :param session: SQLAlchemy Session
    :param policy: Policy
    :param access_result_urls: list of urls generated by access request upload
    :param identity_data: Dict of identity data
    """
    if not identity_data.get(ProvidedIdentityType.email.value):
        raise IdentityNotFoundException(
            "Identity email was not found, so request completion email could not be sent."
        )
    to_identity: Identity = Identity(
        email=identity_data.get(ProvidedIdentityType.email.value),
        phone_number=identity_data.get(ProvidedIdentityType.phone_number.value),
    )
    if policy.get_rules_for_action(action_type=ActionType.access):
        # synchronous for now since failure to send complete emails is fatal to request
        dispatch_message(
            db=session,
            action_type=MessagingActionType.PRIVACY_REQUEST_COMPLETE_ACCESS,
            to_identity=to_identity,
            service_type=CONFIG.notifications.notification_service_type,
            message_body_params=AccessRequestCompleteBodyParams(
                download_links=access_result_urls
            ),
        )
    if policy.get_rules_for_action(action_type=ActionType.erasure):
        dispatch_message(
            db=session,
            action_type=MessagingActionType.PRIVACY_REQUEST_COMPLETE_DELETION,
            to_identity=to_identity,
            service_type=CONFIG.notifications.notification_service_type,
            message_body_params=None,
        )


def initiate_paused_privacy_request_followup(privacy_request: PrivacyRequest) -> None:
    """Initiates scheduler to expire privacy request when the redis cache expires"""
    scheduler.add_job(
        func=mark_paused_privacy_request_as_expired,
        kwargs={"privacy_request_id": privacy_request.id},
        id=privacy_request.id,
        replace_existing=True,
        trigger="date",
        run_date=(datetime.now() + timedelta(seconds=CONFIG.redis.default_ttl_seconds)),
    )


def mark_paused_privacy_request_as_expired(privacy_request_id: str) -> None:
    """Mark "paused" PrivacyRequest as "errored" after its associated identity data in the redis cache has expired."""
    SessionLocal = get_db_session(CONFIG)
    db = SessionLocal()
    privacy_request = PrivacyRequest.get(db=db, object_id=privacy_request_id)
    if not privacy_request:
        logger.info(
            "Attempted to mark as expired. No privacy request with id '{}' found.",
            privacy_request_id,
        )
        db.close()
        return
    if privacy_request.status == PrivacyRequestStatus.paused:
        logger.error(
            "Privacy request '{}' has expired. Please resubmit information.",
            privacy_request.id,
        )
        privacy_request.error_processing(db=db)
    db.close()


def generate_id_verification_code() -> str:
    """
    Generate one-time identity verification code
    """
    return str(random.choice(range(100000, 999999)))


def _retrieve_child_results(  # pylint: disable=R0911
    fides_connector: Tuple[str, ConnectionConfig],
    rule_key: str,
    access_result: Dict[str, List[Row]],
) -> Optional[List[Dict[str, Optional[List[Row]]]]]:
    """Get child access request results to add to upload."""
    try:
        connector = FidesConnector(fides_connector[1])
    except Exception as e:
        logger.error(
            "Error create client for child server {}: {}", fides_connector[0], e
        )
        return None

    results = []

    for key, rows in access_result.items():
        address = CollectionAddress.from_string(key)
        if address.dataset == fides_connector[0]:
            if not rows:
                logger.info("No rows found for result entry {}", key)
                continue
            privacy_request_id = rows[0]["id"]

        if not privacy_request_id:
            logger.error(
                "No privacy request found for connector key {}", fides_connector[0]
            )
            continue

        try:
            client = connector.create_client()
        except requests.exceptions.HTTPError as e:
            logger.error(
                "Error logger into to child server for privacy request {}: {}",
                privacy_request_id,
                e,
            )
            continue

        try:
            request = client.authenticated_request(
                method="get",
                path=f"{V1_URL_PREFIX}{PRIVACY_REQUEST_TRANSFER_TO_PARENT.format(privacy_request_id=privacy_request_id, rule_key=rule_key)}",
                headers={"Authorization": f"Bearer {client.token}"},
            )
            response = client.session.send(request)
        except requests.exceptions.HTTPError as e:
            logger.error(
                "Error retrieving data from child server for privacy request {}: {}",
                privacy_request_id,
                e,
            )
            continue

        if response.status_code != 200:
            logger.error(
                "Error retrieving data from child server for privacy request {}: {}",
                privacy_request_id,
                response.json(),
            )
            continue

        results.append(response.json())

    if not results:
        return None

    return results



from typing import Any, Dict, List, Optional

from pydantic import BaseConfig, create_model
from sqlalchemy import Column, ForeignKey, String, text
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.mutable import MutableList
from sqlalchemy.orm import Session, relationship

from fides.api.ops.models.connectionconfig import ConnectionConfig
from fides.lib.db.base_class import Base
from fides.lib.schemas.base_class import BaseSchema


class AccessManualWebhook(Base):
    """Describes a manual datasource that will be used for access requests.

    These data sources are not treated as part of the traversal.  Data uploaded
    for an AccessManualWebhook is passed on as-is to the end user and is
    not consumed as part of the graph.
    """

    connection_config_id = Column(
        String,
        ForeignKey(ConnectionConfig.id_field_path),
        unique=True,
        nullable=False,
    )
    connection_config = relationship(
        ConnectionConfig, back_populates="access_manual_webhook", uselist=False
    )

    fields = Column(MutableList.as_mutable(JSONB), nullable=False)

    @property
    def fields_schema(self) -> BaseSchema:
        """Build a dynamic Pydantic schema from fields defined on this webhook"""

        class Config:
            extra = "forbid"

        field_definitions: Dict[str, Any] = {
            field["dsr_package_label"]: (Optional[str], None)
            for field in self.fields or []
        }

        ManualWebhookValidationModel = create_model(  # type: ignore
            __model_name="ManualWebhookValidationModel",
            __config__=Config,
            **field_definitions,
        )
        return ManualWebhookValidationModel

    @property
    def fields_non_strict_schema(self) -> BaseSchema:
        """Returns a dynamic Pydantic Schema for webhook fields that can keep the overlap between
        fields that are saved and fields that are defined here."""
        schema: BaseSchema = self.fields_schema
        # Extra is set to "ignore" on the BaseConfig
        schema.__config__ = BaseConfig  # type: ignore[misc]
        return schema

    @property
    def empty_fields_dict(self) -> Dict[str, None]:
        """Return a dictionary that maps defined dsr_package_labels to None

        Returned as a default if no data has been uploaded for a privacy request.
        """
        return {
            key: None
            for key in (self.fields_schema.schema().get("properties") or {}).keys()
        }

    @classmethod
    def get_enabled(cls, db: Session) -> List["AccessManualWebhook"]:
        """Get all enabled access manual webhooks with fields"""
        return (
            db.query(cls)
            .filter(
                AccessManualWebhook.connection_config_id == ConnectionConfig.id,
                ConnectionConfig.disabled.is_(False),
                AccessManualWebhook.fields != text("'null'"),
                AccessManualWebhook.fields != "[]",
            )
            .all()
        )



from __future__ import annotations

from fastapi import HTTPException, status


class AlreadyExistsError(HTTPException):
    """
    To be raised when attempting to create a new resource violates a uniqueness constraint.
    """

    def __init__(self, resource_type: str, fides_key: str) -> None:
        detail = {
            "error": "resource already exists",
            "resource_type": resource_type,
            "fides_key": fides_key,
        }

        super().__init__(status.HTTP_409_CONFLICT, detail=detail)


class DatabaseUnavailableError(HTTPException):
    """
    To be raised when attempting to connect to an uninitialized database.
    """

    def __init__(self, error_message: str = "database unavailable") -> None:
        super().__init__(
            status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail={"error": error_message},
        )


class FidesError(Exception):
    """General fides exception"."""


class NotFoundError(HTTPException):
    """
    To be raised when a requested resource does not exist in the database.
    """

    def __init__(
        self,
        resource_type: str,
        fides_key: str,
        error_message: str = "resource does not exist",
    ) -> None:
        detail = {
            "error": error_message,
            "resource_type": resource_type,
            "fides_key": fides_key,
        }

        super().__init__(status.HTTP_404_NOT_FOUND, detail=detail)


class QueryError(HTTPException):
    """
    To be raised when a database query fails.
    TODO: Improve the messaging here.
    """

    def __init__(self) -> None:
        super().__init__(
            status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail={"error": "a database query failed"},
        )


class ForbiddenError(HTTPException):
    """
    To be raised when a user cannot modify an entity.
    """

    def __init__(self, resource_type: str, fides_key: str) -> None:
        detail = {
            "error": "user does not have permission to modify",
            "resource_type": resource_type,
            "fides_key": fides_key,
        }
        super().__init__(status.HTTP_403_FORBIDDEN, detail=detail)


def get_full_exception_name(exception: Exception) -> str:
    """Get the full exception name
    i.e. get sqlalchemy.exc.IntegrityError instead of just IntegrityError
    """
    module = exception.__class__.__module__
    if module is None or module == str.__class__.__module__:
        return exception.__class__.__name__
    return module + "." + exception.__class__.__name__


class FunctionalityNotConfigured(Exception):
    """Custom exception for when invoked functionality is unavailable due to configuration."""



from enum import Enum
from typing import Any, Dict, Optional

from pydantic import BaseModel


class HTTPMethod(Enum):
    """Enum to represent HTTP Methods"""

    GET = "GET"
    POST = "POST"
    PATCH = "PATCH"
    PUT = "PUT"
    DELETE = "DELETE"
    OPTIONS = "OPTIONS"
    HEAD = "HEAD"


class SaaSRequestParams(BaseModel):
    """
    Holds the method, path, headers, query, and body params to build a SaaS HTTP request.
    """

    method: HTTPMethod
    path: str
    headers: Dict[str, Any] = {}
    query_params: Dict[str, Any] = {}
    body: Optional[str]

    class Config:
        """Using enum values"""

        use_enum_values = True


class ConnectorParamRef(BaseModel):
    """A reference to a value in the connector params (by name)"""

    connector_param: Any


class IdentityParamRef(BaseModel):
    """A reference to the identity type in the filter Post Processor Config"""

    identity: str



from typing import Any, Dict, List

from fastapi import Body, Depends, Security
from fastapi_pagination import Page, Params
from fastapi_pagination.bases import AbstractPage
from fastapi_pagination.ext.sqlalchemy import paginate
from fideslang.validation import FidesKey
from loguru import logger
from pydantic import conlist
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import Session
from starlette.exceptions import HTTPException
from starlette.status import HTTP_200_OK, HTTP_204_NO_CONTENT, HTTP_404_NOT_FOUND

from fides.api.ops.api import deps
from fides.api.ops.api.v1 import scope_registry
from fides.api.ops.api.v1 import urn_registry as urls
from fides.api.ops.common_exceptions import (
    DataCategoryNotSupported,
    DrpActionValidationError,
    PolicyValidationError,
    RuleTargetValidationError,
    RuleValidationError,
)
from fides.api.ops.models.policy import ActionType, Policy, Rule, RuleTarget
from fides.api.ops.models.storage import StorageConfig
from fides.api.ops.schemas import policy as schemas
from fides.api.ops.schemas.api import BulkUpdateFailed
from fides.api.ops.util.api_router import APIRouter
from fides.api.ops.util.logger import Pii
from fides.api.ops.util.oauth_util import verify_oauth_client
from fides.lib.exceptions import KeyOrNameAlreadyExists
from fides.lib.models.client import ClientDetail

router = APIRouter(tags=["DSR Policy"], prefix=urls.V1_URL_PREFIX)


@router.get(
    urls.POLICY_LIST,
    status_code=HTTP_200_OK,
    response_model=Page[schemas.PolicyResponse],
    dependencies=[Security(verify_oauth_client, scopes=[scope_registry.POLICY_READ])],
)
def get_policy_list(
    *,
    db: Session = Depends(deps.get_db),
    params: Params = Depends(),
) -> AbstractPage[Policy]:
    """
    Return a paginated list of all Policy records in this system
    """
    logger.info("Finding all policies with pagination params '{}'", params)
    policies = Policy.query(db=db).order_by(Policy.created_at.desc())
    return paginate(policies, params=params)


def get_policy_or_error(db: Session, policy_key: FidesKey) -> Policy:
    """Helper method to load Policy or throw a 404"""
    logger.info("Finding policy with key '{}'", policy_key)
    policy = Policy.get_by(db=db, field="key", value=policy_key)
    if not policy:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No Policy found for key {policy_key}.",
        )

    return policy


@router.get(
    urls.POLICY_DETAIL,
    status_code=HTTP_200_OK,
    response_model=schemas.PolicyResponse,
    dependencies=[Security(verify_oauth_client, scopes=[scope_registry.POLICY_READ])],
)
def get_policy(
    *,
    policy_key: FidesKey,
    db: Session = Depends(deps.get_db),
) -> schemas.Policy:
    """
    Return a single Policy
    """
    return get_policy_or_error(db, policy_key)  # type: ignore[return-value]


@router.patch(
    urls.POLICY_LIST,
    status_code=HTTP_200_OK,
    response_model=schemas.BulkPutPolicyResponse,
)
def create_or_update_policies(
    *,
    client: ClientDetail = Security(
        verify_oauth_client,
        scopes=[scope_registry.POLICY_CREATE_OR_UPDATE],
    ),
    db: Session = Depends(deps.get_db),
    data: conlist(schemas.Policy, max_items=50) = Body(...),  # type: ignore
) -> schemas.BulkPutPolicyResponse:
    """
    Given a list of policy data elements, create or update corresponding Policy objects
    or report failure
    """
    created_or_updated: List[Policy] = []
    failed: List[BulkUpdateFailed] = []
    logger.info("Starting bulk upsert for {} policies", len(data))

    for policy_schema in data:
        policy_data: Dict[str, Any] = dict(policy_schema)
        try:
            policy = Policy.create_or_update(
                db=db,
                data={
                    "name": policy_data["name"],
                    "key": policy_data.get("key"),
                    "client_id": client.id,
                    "drp_action": policy_data.get("drp_action"),
                    "execution_timeframe": policy_data.get("execution_timeframe"),
                },
            )
        except (
            KeyOrNameAlreadyExists,
            DrpActionValidationError,
            IntegrityError,
        ) as exc:
            logger.warning("Create/update failed for policy: {}", Pii(str(exc)))
            failure = {
                "message": exc.args[0],
                "data": policy_data,
            }
            failed.append(BulkUpdateFailed(**failure))
            continue
        except PolicyValidationError as exc:
            logger.warning("Create/update failed for policy: {}", Pii(str(exc)))
            failure = {
                "message": "This record could not be added because the data provided was invalid.",
                "data": policy_data,
            }
            failed.append(BulkUpdateFailed(**failure))
            continue
        else:
            created_or_updated.append(policy)  # type: ignore[arg-type]

    return schemas.BulkPutPolicyResponse(
        succeeded=created_or_updated,
        failed=failed,
    )


def get_rule_or_error(db: Session, policy_key: FidesKey, rule_key: FidesKey) -> Rule:
    """
    Helper method to load Rule or throw a 404 if it can't be found.
    Also throws a 404 if a `Policy` with the given key can't be found.
    """
    policy = get_policy_or_error(db, policy_key)
    logger.info("Finding rule with key '{}'", rule_key)
    rule = Rule.filter(
        db=db,
        conditions=((Rule.policy_id == policy.id) & (Rule.key == rule_key)),
    ).first()
    if not rule:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No Rule found for key {rule_key} attached to policy {policy_key}",
        )

    return rule


@router.get(
    urls.RULE_LIST,
    status_code=HTTP_200_OK,
    response_model=Page[schemas.RuleResponseWithTargets],
    dependencies=[Security(verify_oauth_client, scopes=[scope_registry.RULE_READ])],
)
def get_rule_list(
    *,
    db: Session = Depends(deps.get_db),
    policy_key: FidesKey,
    params: Params = Depends(),
) -> AbstractPage[Rule]:
    """
    Return a paginated list of all `Rule` records associated with the given `Policy`.
    Throws a 404 if the given `Policy` can't be found.
    """
    policy = get_policy_or_error(db, policy_key)
    logger.info(
        "Finding all rules for policy {} with pagination params '{}'",
        policy_key,
        params,
    )
    rules = Rule.filter(
        db=db,
        conditions=(Rule.policy_id == policy.id),
    ).order_by(Rule.created_at.desc())
    return paginate(rules, params=params)


@router.get(
    urls.RULE_DETAIL,
    status_code=HTTP_200_OK,
    response_model=schemas.RuleResponseWithTargets,
    dependencies=[Security(verify_oauth_client, scopes=[scope_registry.RULE_READ])],
)
def get_rule(
    *,
    policy_key: FidesKey,
    rule_key: FidesKey,
    db: Session = Depends(deps.get_db),
) -> Rule:
    """
    Return a single `Rule` with the given key, associated with the given `Policy`
    """
    return get_rule_or_error(db, policy_key, rule_key)


@router.patch(
    urls.RULE_LIST,
    status_code=HTTP_200_OK,
    response_model=schemas.BulkPutRuleResponse,
)
def create_or_update_rules(
    *,
    client: ClientDetail = Security(
        verify_oauth_client,
        scopes=[scope_registry.RULE_CREATE_OR_UPDATE],
    ),
    policy_key: FidesKey,
    db: Session = Depends(deps.get_db),
    input_data: conlist(schemas.RuleCreate, max_items=50) = Body(...),  # type: ignore
) -> schemas.BulkPutRuleResponse:
    """
    Given a list of Rule data elements, create or update corresponding Rule objects
    or report failure
    """
    logger.info("Finding policy with key '{}'", policy_key)

    policy = get_policy_or_error(db, policy_key)

    created_or_updated: List[Rule] = []
    failed: List[BulkUpdateFailed] = []

    logger.info(
        "Starting bulk upsert for {} rules on policy {}", len(input_data), policy_key
    )

    for schema in input_data:
        # Validate all FKs in the input data exist
        associated_storage_config_id = None
        if schema.action_type == ActionType.access:
            # Only validate the associated StorageConfig on access rules
            storage_destination_key = schema.storage_destination_key
            associated_storage_config: StorageConfig = StorageConfig.get_by(
                db=db,
                field="key",
                value=storage_destination_key,
            )
            if not associated_storage_config:
                logger.warning(
                    "No storage config found with key {}", storage_destination_key
                )
                failure = {
                    "message": f"A StorageConfig with key {storage_destination_key} does not exist",
                    "data": dict(
                        schema
                    ),  # Be sure to pass the schema out the same way it came in
                }
                failed.append(BulkUpdateFailed(**failure))
                continue

            associated_storage_config_id = associated_storage_config.id

        masking_strategy_data = None
        if schema.masking_strategy:
            masking_strategy_data = schema.masking_strategy.dict()

        try:
            rule = Rule.create_or_update(
                db=db,
                data={
                    "action_type": schema.action_type,
                    "client_id": client.id,
                    "key": schema.key,
                    "name": schema.name,
                    "policy_id": policy.id,
                    "storage_destination_id": associated_storage_config_id,
                    "masking_strategy": masking_strategy_data,
                },
            )
        except KeyOrNameAlreadyExists as exc:
            logger.warning(
                "Create/update failed for rule '{}' on policy {}: {}",
                schema.key,
                policy_key,
                exc,
            )
            failure = {
                "message": exc.args[0],
                "data": dict(schema),
            }
            failed.append(BulkUpdateFailed(**failure))
            continue
        except RuleValidationError as exc:
            logger.warning(
                "Create/update failed for rule '{}' on policy {}: {}",
                schema.key,
                policy_key,
                Pii(str(exc)),
            )
            failure = {
                "message": exc.args[0],
                "data": dict(schema),
            }
            failed.append(BulkUpdateFailed(**failure))
            continue
        except ValueError as exc:
            logger.warning(
                "Create/update failed for rule '{}' on policy {}: {}",
                schema.key,
                policy_key,
                Pii(str(exc)),
            )
            failure = {
                "message": exc.args[0],
                "data": dict(schema),
            }
            failed.append(BulkUpdateFailed(**failure))
            continue
        else:
            created_or_updated.append(rule)  # type: ignore[arg-type]

    return schemas.BulkPutRuleResponse(succeeded=created_or_updated, failed=failed)


@router.delete(
    urls.RULE_DETAIL,
    status_code=HTTP_204_NO_CONTENT,
    dependencies=[Security(verify_oauth_client, scopes=[scope_registry.RULE_DELETE])],
)
def delete_rule(
    *,
    policy_key: FidesKey,
    rule_key: FidesKey,
    db: Session = Depends(deps.get_db),
) -> None:
    """
    Delete a policy rule.
    """
    policy = get_policy_or_error(db, policy_key)

    logger.info("Finding rule with key '{}'", rule_key)

    rule = Rule.filter(
        db=db, conditions=(Rule.key == rule_key and Rule.policy_id == policy.id)  # type: ignore[arg-type]
    ).first()
    if not rule:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No Rule found for key {rule_key} on Policy {policy_key}.",
        )

    logger.info("Deleting rule with key '{}'", rule_key)
    rule.delete(db=db)


def get_rule_target_or_error(
    db: Session,
    policy_key: FidesKey,
    rule_key: FidesKey,
    rule_target_key: FidesKey,
) -> RuleTarget:
    """
    Helper method to load Rule Target or throw a 404.
    Also throws a 404 if a `Policy` or `Rule` with the given keys can't be found.
    """
    logger.info("Finding rule target with key '{}'", rule_target_key)
    rule: Rule = get_rule_or_error(db, policy_key, rule_key)
    rule_target = RuleTarget.filter(
        db=db,
        conditions=(
            (RuleTarget.rule_id == rule.id) & (RuleTarget.key == rule_target_key)
        ),
    ).first()
    if not rule_target:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No Rule Target found for key {rule_target_key} attached to rule {rule_key}.",
        )

    return rule_target


@router.get(
    urls.RULE_TARGET_LIST,
    status_code=HTTP_200_OK,
    response_model=Page[schemas.RuleTarget],
    dependencies=[Security(verify_oauth_client, scopes=[scope_registry.RULE_READ])],
)
def get_rule_target_list(
    *,
    db: Session = Depends(deps.get_db),
    policy_key: FidesKey,
    rule_key: FidesKey,
    params: Params = Depends(),
) -> AbstractPage[RuleTarget]:
    """
    Return a paginated list of all `Rule Target` records associated with the given `Rule`
    Throws a 404 if the given `Rule` or `Policy` can't be found.
    """
    rule = get_rule_or_error(db, policy_key, rule_key)
    logger.info(
        "Finding all rule targets for rule {} with pagination params '{}'",
        rule_key,
        params,
    )
    rule_targets = RuleTarget.filter(
        db=db,
        conditions=(RuleTarget.rule_id == rule.id),
    ).order_by(RuleTarget.created_at.desc())
    return paginate(rule_targets, params=params)


@router.get(
    urls.RULE_TARGET_DETAIL,
    status_code=HTTP_200_OK,
    response_model=schemas.RuleTarget,
    dependencies=[Security(verify_oauth_client, scopes=[scope_registry.RULE_READ])],
)
def get_rule_target(
    *,
    policy_key: FidesKey,
    rule_key: FidesKey,
    rule_target_key: FidesKey,
    db: Session = Depends(deps.get_db),
) -> RuleTarget:
    """
    Return a single `Rule Target` associated with the given `Rule` and `Policy`
    """
    return get_rule_target_or_error(db, policy_key, rule_key, rule_target_key)


@router.patch(
    urls.RULE_TARGET_LIST,
    status_code=HTTP_200_OK,
    response_model=schemas.BulkPutRuleTargetResponse,
)
def create_or_update_rule_targets(
    *,
    client: ClientDetail = Security(
        verify_oauth_client, scopes=[scope_registry.RULE_CREATE_OR_UPDATE]
    ),
    policy_key: FidesKey,
    rule_key: FidesKey,
    db: Session = Depends(deps.get_db),
    input_data: conlist(schemas.RuleTarget, max_items=50) = Body(...),  # type: ignore
) -> schemas.BulkPutRuleTargetResponse:
    """
    Given a list of Rule data elements, create corresponding Rule objects
    or report failure
    """
    policy = get_policy_or_error(db, policy_key)

    logger.info("Finding rule with key '{}'", rule_key)
    rule = Rule.filter(
        db=db, conditions=(Rule.key == rule_key and Rule.policy_id == policy.id)  # type: ignore[arg-type]
    ).first()
    if not rule:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No Rule found for key {rule_key} on Policy {policy_key}.",
        )

    created_or_updated = []
    failed = []
    logger.info(
        "Starting bulk upsert for {} rule targets on rule {}", len(input_data), rule_key
    )
    for schema in input_data:
        try:
            target = RuleTarget.create_or_update(
                db=db,
                data={
                    "name": schema.name,
                    "key": schema.key,
                    "data_category": schema.data_category,
                    "rule_id": rule.id,
                    "client_id": client.id,
                },
            )
        except KeyOrNameAlreadyExists as exc:
            logger.warning(
                "Create/update failed for rule target {} on rule {}: {}",
                schema.key,
                rule_key,
                exc,
            )
            failure = {
                "message": exc.args[0],
                "data": dict(schema),
            }
            failed.append(BulkUpdateFailed(**failure))
            continue
        except (
            DataCategoryNotSupported,
            PolicyValidationError,
            RuleTargetValidationError,
        ) as exc:
            logger.warning(
                "Create/update failed for rule target {} on rule {}: {}",
                schema.key,
                rule_key,
                Pii(str(exc)),
            )
            failure = {
                "message": exc.args[0],
                "data": dict(schema),
            }
            failed.append(BulkUpdateFailed(**failure))
            continue
        except IntegrityError as exc:
            logger.warning(
                "Create/update failed for rule target {} on rule {}: {}",
                schema.key,
                rule_key,
                Pii(str(exc)),
            )
            failure = {
                "message": f"DataCategory {schema.data_category} is already specified on Rule with ID {rule.id}",
                "data": dict(schema),
            }
            failed.append(BulkUpdateFailed(**failure))
        else:
            created_or_updated.append(target)

    return schemas.BulkPutRuleTargetResponse(
        succeeded=created_or_updated,
        failed=failed,
    )


@router.delete(
    urls.RULE_TARGET_DETAIL,
    status_code=HTTP_204_NO_CONTENT,
    dependencies=[Security(verify_oauth_client, scopes=[scope_registry.RULE_DELETE])],
)
def delete_rule_target(
    *,
    policy_key: FidesKey,
    rule_key: FidesKey,
    rule_target_key: FidesKey,
    db: Session = Depends(deps.get_db),
) -> None:
    """
    Delete the rule target.
    """
    policy = get_policy_or_error(db, policy_key)

    logger.info("Finding rule with key '{}'", rule_key)
    rule = Rule.filter(
        db=db, conditions=(Rule.key == rule_key and Rule.policy_id == policy.id)  # type: ignore[arg-type]
    ).first()
    if not rule:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No Rule found for key {rule_key} on Policy {policy_key}.",
        )

    logger.info("Finding rule target with key '{}'", rule_target_key)
    target = RuleTarget.filter(
        db=db,
        conditions=(
            RuleTarget.key == rule_target_key and RuleTarget.rule_id == rule.id  # type: ignore[arg-type]
        ),
    ).first()
    if not target:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No RuleTarget found for key {rule_target_key} at Rule {rule_key} on Policy {policy_key}.",
        )

    logger.info("Deleting rule target with key '{}'", rule_target_key)

    target.delete(db=db)



"""
Contains the code that sets up the API.
"""
from datetime import datetime, timezone
from logging import DEBUG, WARNING
from typing import Callable, List, Optional

from fastapi import FastAPI, HTTPException, Request, Response, status
from fastapi.responses import FileResponse
from fideslog.sdk.python.event import AnalyticsEvent
from loguru import logger
from redis.exceptions import RedisError, ResponseError
from slowapi.errors import RateLimitExceeded  # type: ignore
from slowapi.extension import Limiter, _rate_limit_exceeded_handler  # type: ignore
from slowapi.middleware import SlowAPIMiddleware  # type: ignore
from slowapi.util import get_remote_address  # type: ignore
from starlette.background import BackgroundTask
from starlette.middleware.cors import CORSMiddleware
from uvicorn import Config, Server

import fides
from fides.api.ctl import view
from fides.api.ctl.database.database import configure_db
from fides.api.ctl.database.seed import create_or_update_parent_user
from fides.api.ctl.routes import (
    admin,
    crud,
    datamap,
    generate,
    health,
    system,
    validate,
)
from fides.api.ctl.routes.util import API_PREFIX
from fides.api.ctl.ui import (
    get_admin_index_as_response,
    get_local_file_map,
    get_package_file_map,
    get_path_to_admin_ui_file,
    match_route,
    path_is_in_ui_directory,
)
from fides.api.ctl.utils.errors import FidesError
from fides.api.ctl.utils.logger import setup as setup_logging
from fides.api.ops.analytics import (
    accessed_through_local_host,
    in_docker_container,
    send_analytics_event,
)
from fides.api.ops.api.deps import get_api_session
from fides.api.ops.api.v1.api import api_router
from fides.api.ops.api.v1.exception_handlers import ExceptionHandlers
from fides.api.ops.common_exceptions import (
    FunctionalityNotConfigured,
    RedisConnectionError,
)
from fides.api.ops.schemas.analytics import Event, ExtraData
from fides.api.ops.service.connectors.saas.connector_registry_service import (
    load_registry,
    registry_file,
    update_saas_configs,
)

# pylint: disable=wildcard-import, unused-wildcard-import
from fides.api.ops.service.saas_request.override_implementations import *
from fides.api.ops.tasks.scheduled.scheduler import scheduler
from fides.api.ops.util.cache import get_cache
from fides.api.ops.util.logger import _log_exception
from fides.api.ops.util.oauth_util import get_root_client, verify_oauth_client_cli
from fides.core.config import FidesConfig, get_config
from fides.core.config.helpers import check_required_webserver_config_values
from fides.lib.oauth.api.routes.user_endpoints import router as user_router

CONFIG: FidesConfig = get_config()
VERSION = fides.__version__

ROUTERS = crud.routers + [  # type: ignore[attr-defined]
    admin.router,
    datamap.router,
    generate.router,
    health.router,
    validate.router,
    view.router,
    system.router,
]


def create_fides_app(
    cors_origins: List[str] = CONFIG.security.cors_origins,
    routers: List = ROUTERS,
    app_version: str = VERSION,
    api_prefix: str = API_PREFIX,
    request_rate_limit: str = CONFIG.security.request_rate_limit,
    rate_limit_prefix: str = CONFIG.security.rate_limit_prefix,
    security_env: str = CONFIG.security.env.value,
) -> FastAPI:
    """Return a properly configured application."""

    fastapi_app = FastAPI(title="fides", version=app_version)
    fastapi_app.state.limiter = Limiter(
        default_limits=[request_rate_limit],
        headers_enabled=True,
        key_prefix=rate_limit_prefix,
        key_func=get_remote_address,
        retry_after="http-date",
    )
    fastapi_app.add_exception_handler(RateLimitExceeded, _rate_limit_exceeded_handler)
    for handler in ExceptionHandlers.get_handlers():
        fastapi_app.add_exception_handler(FunctionalityNotConfigured, handler)
    fastapi_app.add_middleware(SlowAPIMiddleware)

    if cors_origins:
        fastapi_app.add_middleware(
            CORSMiddleware,
            allow_origins=[str(origin) for origin in cors_origins],
            allow_credentials=True,
            allow_methods=["*"],
            allow_headers=["*"],
        )

    for router in routers:
        fastapi_app.include_router(router)
    fastapi_app.include_router(user_router, tags=["Users"], prefix=f"{api_prefix}")
    fastapi_app.include_router(api_router)

    if security_env == "dev":
        # This removes auth requirements for CLI-related endpoints
        # and is the default
        fastapi_app.dependency_overrides[verify_oauth_client_cli] = get_root_client
    elif security_env == "prod":
        # This is the most secure, so all security deps are maintained
        pass

    return fastapi_app


app = create_fides_app()


@app.middleware("http")
async def dispatch_log_request(request: Request, call_next: Callable) -> Response:
    """
    HTTP Middleware that logs analytics events for each call to Fides endpoints.
    :param request: Request to Fides api
    :param call_next: Callable api endpoint
    :return: Response
    """

    # Only log analytics events for requests that are for API endpoints (i.e. /api/...)
    path = request.url.path
    if (not path.startswith(API_PREFIX)) or (path.endswith("/health")):
        return await call_next(request)

    fides_source: Optional[str] = request.headers.get("X-Fides-Source")
    now: datetime = datetime.now(tz=timezone.utc)
    endpoint = f"{request.method}: {request.url}"

    try:
        response = await call_next(request)
        # HTTPExceptions are considered a handled err by default so are not thrown here.
        # Accepted workaround is to inspect status code of response.
        # More context- https://github.com/tiangolo/fastapi/issues/1840
        response.background = BackgroundTask(
            prepare_and_log_request,
            endpoint,
            request.url.hostname,
            response.status_code,
            now,
            fides_source,
            "HTTPException" if response.status_code >= 400 else None,
        )
        return response

    except Exception as e:
        await prepare_and_log_request(
            endpoint, request.url.hostname, 500, now, fides_source, e.__class__.__name__
        )
        _log_exception(e, CONFIG.dev_mode)
        raise


async def prepare_and_log_request(
    endpoint: str,
    hostname: Optional[str],
    status_code: int,
    event_created_at: datetime,
    fides_source: Optional[str],
    error_class: Optional[str],
) -> None:
    """
    Prepares and sends analytics event provided the user is not opted out of analytics.
    """

    # this check prevents AnalyticsEvent from being called with invalid endpoint during unit tests
    if CONFIG.user.analytics_opt_out:
        return
    await send_analytics_event(
        AnalyticsEvent(
            docker=in_docker_container(),
            event=Event.endpoint_call.value,
            event_created_at=event_created_at,
            local_host=accessed_through_local_host(hostname),
            endpoint=endpoint,
            status_code=status_code,
            error=error_class or None,
            extra_data={ExtraData.fides_source.value: fides_source}
            if fides_source
            else None,
        )
    )


# Configure the routes here so we can generate the openapi json file


@app.on_event("startup")
async def setup_server() -> None:
    "Run all of the required setup steps for the webserver."

    logger.warning(
        "Startup configuration: reloading = {}, dev_mode = {}",
        CONFIG.hot_reloading,
        CONFIG.dev_mode,
    )
    logger.warning("Startup configuration: pii logging = {}", CONFIG.logging.log_pii)

    if CONFIG.logging.level == DEBUG:
        logger.warning(
            "WARNING: log level is DEBUG, so sensitive or personal data may be logged. "
            "Set FIDES__LOGGING__LEVEL to INFO or higher in production."
        )
        CONFIG.log_all_config_values()

    if not CONFIG.database.sync_database_uri:
        raise FidesError("No database uri provided")

    await configure_db(CONFIG.database.sync_database_uri)

    try:
        create_or_update_parent_user()
    except Exception as e:
        logger.error("Error creating parent user: {}", str(e))
        raise FidesError(f"Error creating parent user: {str(e)}")

    logger.info("Validating SaaS connector templates...")
    try:
        registry = load_registry(registry_file)
        db = get_api_session()
        update_saas_configs(registry, db)
        logger.info("Finished loading saas templates")
    except Exception as e:
        logger.error(
            "Error occurred during SaaS connector template validation: {}",
            str(e),
        )
        return
    finally:
        db.close()

    logger.info("Running Cache connection test...")

    try:
        get_cache()
    except (RedisConnectionError, RedisError, ResponseError) as e:
        logger.error("Connection to cache failed: {}", str(e))
        return
    else:
        logger.debug("Connection to cache succeeded")

    if not scheduler.running:
        scheduler.start()

    logger.debug("Sending startup analytics events...")
    await send_analytics_event(
        AnalyticsEvent(
            docker=in_docker_container(),
            event=Event.server_start.value,
            event_created_at=datetime.now(tz=timezone.utc),
        )
    )

    setup_logging(
        CONFIG.logging.level,
        serialize=CONFIG.logging.serialization,
        desination=CONFIG.logging.destination,
    )

    logger.bind(api_config=CONFIG.logging.json()).debug("Configuration options in use")


@app.middleware("http")
async def log_request(request: Request, call_next: Callable) -> Response:
    """Log basic information about every request handled by the server."""
    start = datetime.now()
    response = await call_next(request)
    handler_time = datetime.now() - start
    logger.bind(
        method=request.method,
        status_code=response.status_code,
        handler_time=f"{handler_time.microseconds * 0.001}ms",
        path=request.url.path,
    ).info("Request received")
    return response


# Configure the static file paths last since otherwise it will take over all paths
@app.get("/", tags=["Default"])
def read_index() -> Response:
    """
    Return an index.html at the root path
    """

    return get_admin_index_as_response()


@app.get("/{catchall:path}", response_class=Response, tags=["Default"])
def read_other_paths(request: Request) -> Response:
    """
    Return related frontend files. Adapted from https://github.com/tiangolo/fastapi/issues/130
    """
    # check first if requested file exists (for frontend assets)
    path = request.path_params["catchall"]

    # First search in the local (dev) build for for a matching route.
    ui_file = match_route(get_local_file_map(), path)

    # Next, search for a matching route in the packaged files.
    if not ui_file:
        ui_file = match_route(get_package_file_map(), path)

    # Finally, try to find the exact path as a packaged file.
    if not ui_file:
        ui_file = get_path_to_admin_ui_file(path)

    # Serve up the file as long as it is within the UI directory
    if ui_file and ui_file.is_file():
        if not path_is_in_ui_directory(ui_file):
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND, detail="Item not found"
            )
        logger.debug(
            "catchall request path '{}' matched static admin UI file: {}",
            path,
            ui_file,
        )
        return FileResponse(ui_file)

    # raise 404 for anything that should be backend endpoint but we can't find it
    if path.startswith(API_PREFIX[1:]):
        logger.debug(
            "catchall request path '{}' matched an invalid API route, return 404",
            path,
        )
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Item not found"
        )

    # otherwise return the index
    logger.debug(
        "catchall request path '{}' did not match any admin UI routes, return generic admin UI index",
        path,
    )
    return get_admin_index_as_response()


def start_webserver(port: int = 8080) -> None:
    """Run the webserver."""
    check_required_webserver_config_values()
    server = Server(Config(app, host="0.0.0.0", port=port, log_level=WARNING))

    logger.info(
        "Starting webserver - Host: {}, Port: {}, Log Level: {}",
        server.config.host,
        server.config.port,
        server.config.log_level,
    )
    server.run()



from fides.lib.oauth.scopes import (
    CONNECTION_AUTHORIZE,
    CONNECTION_CREATE_OR_UPDATE,
    CONNECTION_DELETE,
    CONNECTION_READ,
    DATASET_CREATE_OR_UPDATE,
    DATASET_DELETE,
    DATASET_READ,
    PRIVACY_REQUEST_CALLBACK_RESUME,
    PRIVACY_REQUEST_DELETE,
    PRIVACY_REQUEST_READ,
    PRIVACY_REQUEST_REVIEW,
    SAAS_CONFIG_CREATE_OR_UPDATE,
    SAAS_CONFIG_DELETE,
    SAAS_CONFIG_READ,
    USER_CREATE,
    USER_DELETE,
    USER_READ,
)

VIEW = "view"
MANAGE = "manage"

VIEW_SUBJECT_REQUESTS = f"{VIEW}_subject_requests"
MANAGE_SUBJECT_REQUESTS = f"{MANAGE}_subject_requests"

VIEW_CONNECTIONS = f"{VIEW}_datastore_connections"
MANAGE_CONNECTIONS = f"{MANAGE}_datastore_connections"

VIEW_USERS = f"{VIEW}_users"
MANAGE_USERS = f"{MANAGE}_users"


privileges = {
    VIEW_SUBJECT_REQUESTS: {PRIVACY_REQUEST_READ},
    MANAGE_SUBJECT_REQUESTS: {
        PRIVACY_REQUEST_CALLBACK_RESUME,
        PRIVACY_REQUEST_DELETE,
        PRIVACY_REQUEST_READ,
        PRIVACY_REQUEST_REVIEW,
    },
    VIEW_CONNECTIONS: {
        CONNECTION_READ,
        DATASET_READ,
        SAAS_CONFIG_READ,
    },
    MANAGE_CONNECTIONS: {
        CONNECTION_AUTHORIZE,
        CONNECTION_CREATE_OR_UPDATE,
        CONNECTION_DELETE,
        CONNECTION_READ,
        DATASET_CREATE_OR_UPDATE,
        DATASET_DELETE,
        DATASET_READ,
        SAAS_CONFIG_CREATE_OR_UPDATE,
        SAAS_CONFIG_DELETE,
        SAAS_CONFIG_READ,
    },
    VIEW_USERS: {USER_READ},
    MANAGE_USERS: {USER_CREATE, USER_DELETE, USER_READ},
}



from __future__ import annotations

import json
from datetime import datetime
from functools import update_wrapper
from types import FunctionType
from typing import Callable

from fastapi import Depends, HTTPException, Security
from fastapi.security import SecurityScopes
from jose import exceptions
from jose.constants import ALGORITHMS
from loguru import logger
from pydantic import ValidationError
from sqlalchemy.orm import Session
from starlette.status import HTTP_404_NOT_FOUND

from fides.api.ops.api.deps import get_db
from fides.api.ops.api.v1.scope_registry import SCOPE_REGISTRY
from fides.api.ops.api.v1.urn_registry import TOKEN, V1_URL_PREFIX
from fides.api.ops.models.policy import PolicyPreWebhook
from fides.api.ops.schemas.external_https import WebhookJWE
from fides.core.config import get_config
from fides.lib.cryptography.schemas.jwt import (
    JWE_ISSUED_AT,
    JWE_PAYLOAD_CLIENT_ID,
    JWE_PAYLOAD_SCOPES,
)
from fides.lib.exceptions import AuthenticationError, AuthorizationError
from fides.lib.models.client import ClientDetail
from fides.lib.models.fides_user import FidesUser
from fides.lib.oauth.oauth_util import extract_payload, is_token_expired
from fides.lib.oauth.schemas.oauth import OAuth2ClientCredentialsBearer

CONFIG = get_config()
JWT_ENCRYPTION_ALGORITHM = ALGORITHMS.A256GCM


# TODO: include list of all scopes in the docs via the scopes={} dict
# (see https://fastapi.tiangolo.com/advanced/security/oauth2-scopes/)
oauth2_scheme = OAuth2ClientCredentialsBearer(
    tokenUrl=(V1_URL_PREFIX + TOKEN),
)


def copy_func(source_function: Callable) -> Callable:
    """Based on http://stackoverflow.com/a/6528148/190597 (Glenn Maynard)"""
    target_function = FunctionType(
        source_function.__code__,
        source_function.__globals__,
        name=source_function.__name__,
        argdefs=source_function.__defaults__,
        closure=source_function.__closure__,
    )
    target_function = update_wrapper(target_function, source_function)
    target_function.__kwdefaults__ = source_function.__kwdefaults__
    return target_function


async def get_current_user(
    security_scopes: SecurityScopes,
    authorization: str = Security(oauth2_scheme),
    db: Session = Depends(get_db),
) -> FidesUser:
    """A wrapper around verify_oauth_client that returns that client's user if one exists."""
    client = await verify_oauth_client(
        security_scopes=security_scopes,
        authorization=authorization,
        db=db,
    )

    if client.id == CONFIG.security.oauth_root_client_id:
        return FidesUser(
            id=CONFIG.security.oauth_root_client_id,
            username=CONFIG.security.root_username,
            created_at=datetime.utcnow(),
        )

    return client.user  # type: ignore[attr-defined]


def is_callback_token_expired(issued_at: datetime | None) -> bool:
    """Returns True if the token is older than the expiration of the redis cache.  We
    can't resume executing the privacy request if the identity data is gone.
    """
    if not issued_at:
        return True

    return (
        datetime.now() - issued_at
    ).total_seconds() / 60.0 > CONFIG.execution.privacy_request_delay_timeout


def verify_callback_oauth(
    security_scopes: SecurityScopes,
    authorization: str = Security(oauth2_scheme),
    db: Session = Depends(get_db),
) -> PolicyPreWebhook:
    """
    Verifies the specific token that accompanies a request when a user wants to resume executing
    a PrivacyRequest after it was paused by a webhook. Note that this token was sent along with the
    request when calling the webhook originally.
    Verifies that the webhook token hasn't expired and loads the webhook from that token.
    Also verifies scopes, but note that this was given to the user in a request header and they've
    just returned it back.
    """
    if authorization is None:
        raise AuthenticationError(detail="Authentication Failure")

    token_data = json.loads(
        extract_payload(authorization, CONFIG.security.app_encryption_key)
    )
    try:
        token = WebhookJWE(**token_data)
    except ValidationError:
        raise AuthorizationError(detail="Not Authorized for this action")

    assigned_scopes = token_data[JWE_PAYLOAD_SCOPES]
    if not set(security_scopes.scopes).issubset(assigned_scopes):
        raise AuthorizationError(detail="Not Authorized for this action")

    if is_callback_token_expired(datetime.fromisoformat(token.iat)):
        raise AuthorizationError(detail="Webhook token expired")

    webhook = PolicyPreWebhook.get_by(db, field="id", value=token.webhook_id)

    if not webhook:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No Policy Pre-Execution Webhook found with id '{token.webhook_id}'.",
        )
    return webhook


async def get_root_client(
    db: Session = Depends(get_db), client_id: str = CONFIG.security.oauth_root_client_id
) -> ClientDetail:
    """
    Gets the root_client directly.

    This function is primarily used to let users bypass endpoint authorization
    """
    client = ClientDetail.get(
        db, object_id=client_id, config=CONFIG, scopes=SCOPE_REGISTRY
    )
    if not client:
        logger.debug("Auth token belongs to an invalid client_id.")
        raise AuthorizationError(detail="Not Authorized for this action")
    return client


async def verify_oauth_client(
    security_scopes: SecurityScopes,
    authorization: str = Security(oauth2_scheme),
    db: Session = Depends(get_db),
) -> ClientDetail:
    """
    Verifies that the access token provided in the authorization header contains
    the necessary scopes specified by the caller. Yields a 403 forbidden error
    if not.

    NOTE: This function may be overwritten in `main.py` when changing
    the security environment.
    """
    if authorization is None:
        logger.debug("No authorization supplied.")
        raise AuthenticationError(detail="Authentication Failure")

    try:
        token_data = json.loads(
            extract_payload(authorization, CONFIG.security.app_encryption_key)
        )
    except exceptions.JWEParseError as exc:
        logger.debug("Unable to parse auth token.")
        raise AuthorizationError(detail="Not Authorized for this action") from exc

    issued_at = token_data.get(JWE_ISSUED_AT, None)
    if not issued_at:
        logger.debug("Auth token expired.")
        raise AuthorizationError(detail="Not Authorized for this action")

    if is_token_expired(
        datetime.fromisoformat(issued_at),
        CONFIG.security.oauth_access_token_expire_minutes,
    ):
        raise AuthorizationError(detail="Not Authorized for this action")

    assigned_scopes = token_data[JWE_PAYLOAD_SCOPES]
    if not set(security_scopes.scopes).issubset(assigned_scopes):
        scopes_required = ",".join(security_scopes.scopes)
        scopes_provided = ",".join(assigned_scopes)
        logger.debug(
            "Auth token missing required scopes: {}. Scopes provided: {}.",
            scopes_required,
            scopes_provided,
        )
        raise AuthorizationError(detail="Not Authorized for this action")

    client_id = token_data.get(JWE_PAYLOAD_CLIENT_ID)
    if not client_id:
        logger.debug("No client_id included in auth token.")
        raise AuthorizationError(detail="Not Authorized for this action")

    # scopes param is only used if client is root client, otherwise we use the client's associated scopes
    client = ClientDetail.get(
        db, object_id=client_id, config=CONFIG, scopes=SCOPE_REGISTRY
    )

    if not client:
        logger.debug("Auth token belongs to an invalid client_id.")
        raise AuthorizationError(detail="Not Authorized for this action")

    if not set(assigned_scopes).issubset(set(client.scopes)):
        # If the scopes on the token are not a subset of the scopes available
        # to the associated oauth client, this token is not valid
        logger.debug("Client no longer allowed to issue these scopes.")
        raise AuthorizationError(detail="Not Authorized for this action")
    return client


# This is a workaround so that we can override CLI-related endpoints and
# all other endpoints separately.
verify_oauth_client_cli = copy_func(verify_oauth_client)



from typing import List, Optional
from urllib.parse import urlencode
from uuid import uuid4

from requests import PreparedRequest
from sqlalchemy.orm import Session

from fides.api.ops.common_exceptions import FidesopsException
from fides.api.ops.models.authentication_request import AuthenticationRequest
from fides.api.ops.models.connectionconfig import ConnectionConfig
from fides.api.ops.schemas.saas.strategy_configuration import (
    OAuth2AuthorizationCodeConfiguration,
)
from fides.api.ops.service.authentication.authentication_strategy_oauth2_base import (
    OAuth2AuthenticationStrategyBase,
)
from fides.api.ops.util.saas_util import assign_placeholders, map_param_values
from fides.core.config import get_config

CONFIG = get_config()


class OAuth2AuthorizationCodeAuthenticationStrategy(OAuth2AuthenticationStrategyBase):
    """
    Checks the expiration date on the stored access token and refreshes
    it if needed using the configured token refresh request.
    """

    name = "oauth2_authorization_code"
    configuration_model = OAuth2AuthorizationCodeConfiguration

    def __init__(self, configuration: OAuth2AuthorizationCodeConfiguration):
        super().__init__(configuration)
        self.authorization_request = configuration.authorization_request

    def add_authentication(
        self, request: PreparedRequest, connection_config: ConnectionConfig
    ) -> PreparedRequest:
        """
        Checks the expiration date on the existing access token and refreshes if necessary.
        The existing/updated access token is then added to the request as a bearer token.
        """

        # make sure required secrets have been provided
        self._check_required_secrets(connection_config)

        access_token = connection_config.secrets.get("access_token")  # type: ignore
        if not access_token:
            raise FidesopsException(
                f"OAuth2 access token not found for {connection_config.key}, please "
                f"authenticate connection via /api/v1/connection/{connection_config.key}/authorize"
            )

        # The refresh request is optional since the OAuth2 spec does not require
        # a refresh token to be returned as part of an access token request
        #
        # https://datatracker.ietf.org/doc/html/rfc6749#section-5.1

        access_token = self._refresh_token(connection_config)

        # add access_token to request
        request.headers["Authorization"] = "Bearer " + access_token
        return request

    @property
    def _required_secrets(self) -> List[str]:
        return ["client_id", "client_secret", "redirect_uri"]

    def get_authorization_url(
        self, db: Session, connection_config: ConnectionConfig
    ) -> Optional[str]:
        """
        Returns the authorization URL to initiate the OAuth2 workflow.
        Also stores a reference between the authorization request and the connector
        to be able to link the returned auth code to the correct connector."""

        # make sure the required secrets to generate the authorization url have been provided
        self._check_required_secrets(connection_config)

        # generate the state that will be used to link this authorization request to this connector
        state = self._generate_state()
        AuthenticationRequest.create_or_update(
            db, data={"connection_key": connection_config.key, "state": state}
        )
        # add state to secrets
        connection_config.secrets["state"] = state  # type: ignore

        # assign placeholders in the authorization request config
        prepared_authorization_request = map_param_values(
            "authorize",
            f"{connection_config.name} OAuth2",
            self.authorization_request,
            connection_config.secrets,  # type: ignore
        )

        # get the client config from the authorization request or default
        # to the base client config if one isn't provided
        client_config = (
            self.authorization_request.client_config
            if self.authorization_request.client_config
            else connection_config.get_saas_config().client_config  # type: ignore
        )

        # build the complete URL with query params
        return (
            f"{client_config.protocol}://{assign_placeholders(client_config.host, connection_config.secrets)}"  # type: ignore
            f"{prepared_authorization_request.path}"
            f"?{urlencode(prepared_authorization_request.query_params)}"
        )

    @staticmethod
    def _generate_state() -> str:
        """
        Generates a string value to associate the authentication request
        with an eventual OAuth2 callback response.

        If an oauth_instance name is defined then the name is added as a
        prefix to the generated state. The state prefix can be used by a
        proxy server to route the callback response to a specific Fidesops
        instance. This functionality is not part of the OAuth2 specification
        but it can be used for local testing of OAuth2 workflows.
        """

        state = str(uuid4())
        if CONFIG.oauth_instance:
            state = f"{CONFIG.oauth_instance}-{state}"
        return state



from datetime import datetime
from enum import Enum as EnumType
from typing import Any, Dict, List, Optional, Union

from fideslang.validation import FidesKey
from pydantic import Field, validator

from fides.api.ops.models.policy import ActionType
from fides.api.ops.models.privacy_request import (
    CheckpointActionRequired,
    ExecutionLogStatus,
    PrivacyRequestStatus,
)
from fides.api.ops.schemas.api import BulkResponse, BulkUpdateFailed
from fides.api.ops.schemas.base_class import BaseSchema
from fides.api.ops.schemas.policy import PolicyResponse as PolicySchema
from fides.api.ops.schemas.redis_cache import Identity
from fides.api.ops.util.encryption.aes_gcm_encryption_scheme import (
    verify_encryption_key,
)
from fides.core.config import get_config
from fides.lib.models.audit_log import AuditLogAction
from fides.lib.oauth.schemas.user import PrivacyRequestReviewer

CONFIG = get_config()


class PrivacyRequestDRPStatus(EnumType):
    """A list of privacy request statuses specified by the Data Rights Protocol."""

    open = "open"
    in_progress = "in_progress"
    fulfilled = "fulfilled"
    revoked = "revoked"
    denied = "denied"
    expired = "expired"


class PrivacyRequestDRPStatusResponse(BaseSchema):
    """A Fidesops PrivacyRequest updated to fit the Data Rights Protocol specification."""

    request_id: str
    received_at: datetime
    expected_by: Optional[datetime]
    processing_details: Optional[str]
    status: PrivacyRequestDRPStatus
    reason: Optional[str]
    user_verification_url: Optional[str]

    class Config:
        """Set orm_mode and use_enum_values"""

        orm_mode = True
        use_enum_values = True


class Consent(BaseSchema):
    """Schema for consent."""

    data_use: str
    data_use_description: Optional[str] = None
    opt_in: bool


class PrivacyRequestCreate(BaseSchema):
    """Data required to create a PrivacyRequest"""

    external_id: Optional[str]
    started_processing_at: Optional[datetime]
    finished_processing_at: Optional[datetime]
    requested_at: Optional[datetime]
    identity: Identity
    policy_key: FidesKey
    encryption_key: Optional[str] = None
    consent_preferences: Optional[List[Consent]] = None

    @validator("encryption_key")
    def validate_encryption_key(
        cls: "PrivacyRequestCreate", value: Optional[str] = None
    ) -> Optional[str]:
        """Validate encryption key where applicable"""
        if value:
            verify_encryption_key(value.encode(CONFIG.security.encoding))
        return value


class FieldsAffectedResponse(BaseSchema):
    """Schema detailing the individual fields affected by a particular query detailed in the ExecutionLog"""

    path: Optional[str]
    field_name: Optional[str]
    data_categories: Optional[List[str]]

    class Config:
        """Set orm_mode and use_enum_values"""

        orm_mode = True
        use_enum_values = True


class ExecutionLogResponse(BaseSchema):
    """Schema for the embedded ExecutionLogs associated with a PrivacyRequest"""

    collection_name: Optional[str]
    fields_affected: Optional[List[FieldsAffectedResponse]]
    message: Optional[str]
    action_type: ActionType
    status: ExecutionLogStatus
    updated_at: Optional[datetime]

    class Config:
        """Set orm_mode and use_enum_values"""

        orm_mode = True
        use_enum_values = True


class ExecutionLogDetailResponse(ExecutionLogResponse):
    """Schema for the detailed ExecutionLogs when accessed directly"""

    connection_key: Optional[str]
    dataset_name: Optional[str]


class ExecutionAndAuditLogResponse(BaseSchema):
    """Schema for the combined ExecutionLogs and Audit Logs
    associated with a PrivacyRequest"""

    connection_key: Optional[str]
    collection_name: Optional[str]
    fields_affected: Optional[List[FieldsAffectedResponse]]
    message: Optional[str]
    action_type: Optional[ActionType]
    status: Optional[Union[ExecutionLogStatus, AuditLogAction]]
    updated_at: Optional[datetime]
    user_id: Optional[str]

    class Config:
        """Set orm_mode and allow population by field name"""

        use_enum_values = True
        allow_population_by_field_name = True


class RowCountRequest(BaseSchema):
    """Schema for a user to manually confirm data erased for a collection"""

    row_count: int


class CheckpointActionRequiredDetails(CheckpointActionRequired):
    collection: Optional[str] = None  # type: ignore


class VerificationCode(BaseSchema):
    """Request Body for the user to supply their identity verification code"""

    code: str


class ManualWebhookData(BaseSchema):
    checked: bool  # If we have record the user saved data for this webhook (even if it was empty)
    fields: Dict[str, Any]


class PrivacyRequestNotificationInfo(BaseSchema):
    email_addresses: List[str]
    notify_after_failures: int


class PrivacyRequestResponse(BaseSchema):
    """Schema to check the status of a PrivacyRequest"""

    id: str
    created_at: Optional[datetime]
    started_processing_at: Optional[datetime]
    reviewed_at: Optional[datetime]
    reviewed_by: Optional[str]
    reviewer: Optional[PrivacyRequestReviewer]
    finished_processing_at: Optional[datetime]
    identity_verified_at: Optional[datetime]
    paused_at: Optional[datetime]
    status: PrivacyRequestStatus
    external_id: Optional[str]
    # This field intentionally doesn't use the Identity schema
    # as it is an API response field, and we don't want to reveal any more
    # about our PII structure than is explicitly stored in the cache on request
    # creation.
    identity: Optional[Dict[str, Optional[str]]]
    policy: PolicySchema
    action_required_details: Optional[CheckpointActionRequiredDetails] = None
    resume_endpoint: Optional[str]
    days_left: Optional[int]

    class Config:
        """Set orm_mode and use_enum_values"""

        orm_mode = True
        use_enum_values = True


class PrivacyRequestVerboseResponse(PrivacyRequestResponse):
    """The schema for the more detailed PrivacyRequest response containing both
    detailed execution logs and audit logs."""

    execution_and_audit_logs_by_dataset: Dict[
        str, List[ExecutionAndAuditLogResponse]
    ] = Field(alias="results")

    class Config:
        """Allow the results field to be populated by the 'PrivacyRequest.execution_logs_by_dataset' property"""

        allow_population_by_field_name = True


class ReviewPrivacyRequestIds(BaseSchema):
    """Pass in a list of privacy request ids"""

    request_ids: List[str] = Field(..., max_items=50)


class DenyPrivacyRequests(ReviewPrivacyRequestIds):
    """Pass in a list of privacy request ids and rejection reason"""

    reason: Optional[str]


class BulkPostPrivacyRequests(BulkResponse):
    """Schema with mixed success/failure responses for Bulk Create of PrivacyRequest responses."""

    succeeded: List[PrivacyRequestResponse]
    failed: List[BulkUpdateFailed]


class BulkReviewResponse(BulkPostPrivacyRequests):
    """Schema with mixed success/failure responses for Bulk Approve/Deny of PrivacyRequest responses."""


class ConsentPreferences(BaseSchema):
    """Schema for consent preferences."""

    consent: Optional[List[Consent]] = None


class ConsentWithExecutableStatus(BaseSchema):
    """Schema for executable consents"""

    data_use: str
    executable: bool


class ConsentPreferencesWithVerificationCode(BaseSchema):
    """Schema for consent preferences including the verification code."""

    code: Optional[str]
    consent: List[Consent]
    policy_key: Optional[FidesKey] = None
    executable_options: Optional[List[ConsentWithExecutableStatus]]
    browser_identity: Optional[Identity]


class ConsentRequestResponse(BaseSchema):
    """Schema for consent request response."""

    consent_request_id: str


class ConsentRequestVerification(BaseSchema):
    """Schema for consent requests."""

    identity: Identity
    data_use: str
    data_use_description: Optional[str] = None
    opt_in: bool



"""migrate to usage of evaluation violations

Revision ID: 312aff72b275
Revises: 45c7a349db68
Create Date: 2021-12-15 15:53:46.484775

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "312aff72b275"
down_revision = "45c7a349db68"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("evaluations", sa.Column("violations", sa.JSON(), nullable=True))
    op.drop_column("evaluations", "details")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "evaluations",
        sa.Column(
            "details",
            postgresql.ARRAY(sa.VARCHAR()),
            autoincrement=False,
            nullable=True,
        ),
    )
    op.drop_column("evaluations", "violations")
    # ### end Alembic commands ###
